# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module for downloading and curating data from ECMWFs ERA5 dataset (via CDS).

For further reference see
https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation
"""

import os
import numpy as np
import xarray as xr
from tempfile import mkstemp
import weakref
import cdsapi
import logging
from numpy import atleast_1d
from ..gis import maybe_swap_spatial_dims

# Null context for running a with statements wihout any context
try:
    from contextlib import nullcontext
except ImportError:
    # for Python verions < 3.7:
    import contextlib
    @contextlib.contextmanager
    def nullcontext():
        yield

logger = logging.getLogger(__name__)

# Model and CRS Settings
crs = 4326

features = {
    'height': ['height'],
    'wind': [
        'wnd100m',
        'roughness'],
    'influx': [
        'influx_toa',
        'influx_direct',
        'influx_diffuse',
        'albedo'],
    'temperature': [
        'temperature',
        'soil temperature'],
    'runoff': ['runoff']}

static_features = {'height'}


def _add_height(ds):
    """Convert geopotential 'z' to geopotential height following [1].

    References
    ----------
    [1] ERA5: surface elevation and orography, retrieved: 10.02.2019
    https://confluence.ecmwf.int/display/CKB/ERA5%3A+surface+elevation+and+orography

    """
    g0 = 9.80665
    z = ds['z']
    if 'time' in z.coords:
        z = z.isel(time=0, drop=True)
    ds['height'] = z / g0
    ds = ds.drop_vars('z')
    return ds



def _rename_and_clean_coords(ds, add_lon_lat=True):
    """Rename 'longitude' and 'latitude' columns to 'x' and 'y' and fix roundings.

    Optionally (add_lon_lat, default:True) preserves latitude and longitude
    columns as 'lat' and 'lon'.
    """
    ds = ds.rename({'longitude': 'x', 'latitude': 'y'})
    # round coords since cds coords are float32 which would lead to mismatches
    ds = ds.assign_coords(x=np.round(ds.x.astype(float), 5),
                          y=np.round(ds.y.astype(float), 5))
    ds = maybe_swap_spatial_dims(ds)
    if add_lon_lat:
        ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])
    return ds


def get_data_wind(retrieval_params):
    """Get wind data for given retrieval parameters."""
    ds = retrieve_data(
        variable=[
            '100m_u_component_of_wind',
            '100m_v_component_of_wind',
            'forecast_surface_roughness'],
        **retrieval_params)
    ds = _rename_and_clean_coords(ds)

    ds['wnd100m'] = (np.sqrt(ds['u100']**2 + ds['v100']**2)
                     .assign_attrs(units=ds['u100'].attrs['units'],
                                   long_name="100 metre wind speed"))
    ds = ds.drop_vars(['u100', 'v100'])
    ds = ds.rename({'fsr': 'roughness'})

    return ds


def sanitize_wind(ds):
    """Sanitize retrieved wind data."""
    ds['roughness'] = ds['roughness'].where(ds['roughness'] >= 0.0, 2e-4)
    return ds


def get_data_influx(retrieval_params):
    """Get influx data for given retrieval parameters."""
    ds = retrieve_data(
        variable=[
            'surface_net_solar_radiation',
            'surface_solar_radiation_downwards',
            'toa_incident_solar_radiation',
            'total_sky_direct_solar_radiation_at_surface'],
        **retrieval_params)

    ds = _rename_and_clean_coords(ds)

    ds = ds.rename({'fdir': 'influx_direct', 'tisr': 'influx_toa'})
    ds['albedo'] = (((ds['ssrd'] - ds['ssr']) /
                     ds['ssrd'].where(ds['ssrd'] != 0))
                    .fillna(0.)
                    .assign_attrs(units='(0 - 1)', long_name='Albedo'))
    ds['influx_diffuse'] = (
        (ds['ssrd'] - ds['influx_direct'])
        .assign_attrs(units='J m**-2',
                      long_name='Surface diffuse solar radiation downwards'))
    ds = ds.drop_vars(['ssrd', 'ssr'])

    # Convert from energy to power J m**-2 -> W m**-2 and clip negative fluxes
    for a in ('influx_direct', 'influx_diffuse', 'influx_toa'):
        ds[a] = ds[a] / (60. * 60.)
        ds[a].attrs['units'] = 'W m**-2'

    return ds


def sanitize_inflow(ds):
    """Sanitize retrieved inflow data."""
    for a in ('influx_direct', 'influx_diffuse', 'influx_toa'):
        ds[a] = ds[a].clip(min=0.)
    return ds


def get_data_temperature(retrieval_params):
    """Get wind temperature for given retrieval parameters."""
    ds = retrieve_data(variable=['2m_temperature', 'soil_temperature_level_4'],
                       **retrieval_params)

    ds = _rename_and_clean_coords(ds)
    ds = ds.rename({'t2m': 'temperature', 'stl4': 'soil temperature'})

    return ds


def get_data_runoff(retrieval_params):
    """Get runoff data for given retrieval parameters."""
    ds = retrieve_data(variable=['runoff'], **retrieval_params)

    ds = _rename_and_clean_coords(ds)
    ds = ds.rename({'ro': 'runoff'})

    return ds


def sanitize_runoff(ds):
    """Sanitize retrieved runoff data."""
    ds['runoff'] = ds['runoff'].clip(min=0.)
    return ds


def get_data_height(retrieval_params):
    """Get height data for given retrieval parameters."""
    ds = retrieve_data(variable='orography', **retrieval_params)

    ds = _rename_and_clean_coords(ds)
    ds = _add_height(ds)

    return ds


def _area(coords):
    # North, West, South, East. Default: global
    x0, x1 = coords['x'].min().item(), coords['x'].max().item()
    y0, y1 = coords['y'].min().item(), coords['y'].max().item()
    return [y1, x0, y0, x1]


def retrieval_times(coords, static=False):
    """
    Get list of retrieval cdsapi arguments for time dimension in coordinates.

    If static is False, this function creates a query for each year in the
    time axis in coords. This ensures not running into query limits of the
    cdsapi. If static is True, the function return only one set of parameters
    for the very first time point.

    Parameters
    ----------
    coords : atlite.Cutout.coords

    Returns
    -------
    list of dicts witht retrieval arguments

    """
    time = coords['time'].to_index()
    if static:
        return {'year': str(time[0].year), 'month': str(time[0].month),
                'day': str(time[0].day), 'time': time[0].strftime("%H:00")}

    times = []
    for year in time.year.unique():
        t = time[time.year==year]
        query = {'year': str(year),
                 'month': list(t.month.unique()),
                 'day': list(t.day.unique()),
                 'time': ["%02d:00" %h for h in t.hour.unique()]}
        times.append(query)
    return times


def noisy_unlink(path):
    """Delete file at given path."""
    logger.debug(f"Deleting file {path}")
    try:
        os.unlink(path)
    except PermissionError:
        logger.error(f"Unable to delete file {path}, as it is still in use.")


def retrieve_data(product, chunks=None, tmpdir=None, lock=None, **updates):
    """
    Download data like ERA5 from the Climate Data Store (CDS).

    If you want to track the state of your request go to
    https://cds.climate.copernicus.eu/cdsapp#!/yourrequests
    """
    request = {
        'product_type': 'reanalysis',
        'format': 'netcdf'}
    request.update(updates)

    assert {'year', 'month', 'variable'}.issubset(request), (
            "Need to specify at least 'variable', 'year' and 'month'")

    client = cdsapi.Client(info_callback=logger.debug,
                            debug=logging.DEBUG >= logging.root.level)
    result = client.retrieve(product, request)

    if lock is None:
        lock = nullcontext()

    with lock:
        fd, target = mkstemp(suffix='.nc', dir=tmpdir); os.close(fd)

        yearstr = ', '.join(atleast_1d(request['year']))
        variables = atleast_1d(request['variable'])
        varstr = ''.join(['\t * ' + v + f' ({yearstr})\n' for v in variables])
        logger.info(f"CDS: Downloading variables\n{varstr}")
        result.download(target)

    ds = xr.open_dataset(target, chunks=chunks or {})
    if tmpdir is None:
        logger.debug(f"Adding finalizer for {target}")
        weakref.finalize(ds._file_obj._manager, noisy_unlink, target)

    return ds



def get_data(cutout, feature, tmpdir, lock=None, **creation_parameters):
    """
    Retrieve data from ECMWFs ERA5 dataset (via CDS).

    This front-end function downloads data for a specific feature and formats
    it to match the given Cutout.

    Parameters
    ----------
    cutout : atlite.Cutout
    feature : str
        Name of the feature data to retrieve. Must be in
        `atlite.datasets.era5.features`
    tmpdir : str/Path
        Directory where the temporary netcdf files are stored.
    **creation_parameters :
        Additional keyword arguments. The only effective argument is 'sanitize'
        (default True) which sets sanitization of the data on or off.

    Returns
    -------
    xarray.Dataset
        Dataset of dask arrays of the retrieved variables.

    """
    coords = cutout.coords

    sanitize = creation_parameters.get('sanitize', True)

    retrieval_params = {'product': 'reanalysis-era5-single-levels',
                        'area': _area(coords),
                        'chunks': cutout.chunks,
                        'grid': [cutout.dx, cutout.dy],
                        'tmpdir': tmpdir,
                        'lock': lock}

    func = globals().get(f"get_data_{feature}")
    sanitize_func = globals().get(f"sanitize_{feature}")

    logger.info(f'Requesting data for feature {feature}...')

    def retrieve_once(time):
        ds = func({**retrieval_params, **time})
        if sanitize and sanitize_func is not None:
            ds = sanitize_func(ds)
        return ds

    if feature in static_features:
        return retrieve_once(retrieval_times(coords, static=True)).squeeze()

    datasets = map(retrieve_once, retrieval_times(coords))

    return xr.concat(datasets, dim='time').sel(time=coords['time'])



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Management of data retrieval and structure.
"""

import pandas as pd
import xarray as xr
import os
from numpy import atleast_1d
from tempfile import mkstemp, mkdtemp
from shutil import rmtree
from functools import wraps
from dask import delayed, compute
from dask.utils import SerializableLock
from dask.diagnostics import ProgressBar
import logging
logger = logging.getLogger(__name__)

from .datasets import modules as datamodules


def get_features(cutout, module, features, tmpdir=None):
    """
    Load the feature data for a given module.

    This get the data for a set of features from a module. All modules in
    `atlite.datasets` are allowed.
    """
    parameters = cutout.data.attrs
    lock = SerializableLock()
    datasets = []
    get_data = datamodules[module].get_data

    for feature in features:
        feature_data = delayed(get_data)(cutout, feature, tmpdir=tmpdir,
                                         lock=lock, **parameters)
        datasets.append(feature_data)

    datasets = compute(*datasets)

    ds = xr.merge(datasets, compat='equals')
    for v in ds:
        ds[v].attrs['module'] = module
        fd = datamodules[module].features.items()
        ds[v].attrs['feature'] = [k for k, l in fd if v in l].pop()
    return ds


def available_features(module=None):
    """
    Inspect the available features of all or a selection of modules.

    Parameters
    ----------
    module : str/list, optional
        Module name(s) which to inspect. The default None will result in all
        modules

    Returns
    -------
    pd.Series
        A Series of all variables. The MultiIndex indicated which module
        provides the variable and with which feature name the variable can be
        obtained.

    """
    features = {name: m.features for name, m in datamodules.items()}
    features =  pd.DataFrame(features).unstack().dropna() \
                  .rename_axis(index=['module', 'feature']).rename('variables')
    if module is not None:
        features = features.reindex(atleast_1d(module), level='module')
    return features.explode()


def non_bool_dict(d):
    """Convert bool to int for netCDF4 storing"""
    return {k: v if not isinstance(v, bool) else int(v) for k,v in d.items()}


def maybe_remove_tmpdir(func):
    'Use this wrapper to make tempfile deletion compatible with windows machines.'
    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get('tmpdir', None):
            res = func(*args, **kwargs)
        else:
            kwargs['tmpdir'] = mkdtemp()
            try:
                res = func(*args, **kwargs)
            finally:
                rmtree(kwargs['tmpdir'])
        return res
    return wrapper


@maybe_remove_tmpdir
def cutout_prepare(cutout, features=None, tmpdir=None, overwrite=False):
    """
    Prepare all or a selection of features in a cutout.

    This function loads the feature data of a cutout, e.g. influx or runoff.
    When not specifying the `feature` argument, all available features will be
    loaded. The function compares the variables which are already included in
    the cutout with the available variables of the modules specified by the
    cutout. It detects missing variables and stores them into the netcdf file
    of the cutout.


    Parameters
    ----------
    cutout : atlite.Cutout
    features : str/list, optional
        Feature(s) to be prepared. The default slice(None) results in all
        available features.
    tmpdir : str/Path, optional
        Directory in which temporary files (for example retrieved ERA5 netcdf
        files) are stored. If set, the directory will not be deleted and the
        intermediate files can be examined.
    overwrite : bool, optional
        Whether to overwrite variables which are already included in the
        cutout. The default is False.

    Returns
    -------
    cutout : atlite.Cutout
        Cutout with prepared data. The variables are stored in `cutout.data`.

    """
    if cutout.prepared and not overwrite:
        logger.info('Cutout already prepared.')
        return cutout

    logger.info(f'Storing temporary files in {tmpdir}')

    modules = atleast_1d(cutout.module)
    features = atleast_1d(features) if features else slice(None)
    prepared = set(atleast_1d(cutout.data.attrs['prepared_features']))

    # target is series of all available variables for given module and features
    target = available_features(modules).loc[:, features].drop_duplicates()

    for module in target.index.unique('module'):
        missing_vars = target[module]
        if not overwrite:
            missing_vars = missing_vars[lambda v: ~v.isin(cutout.data)]
        if missing_vars.empty:
            continue
        logger.info(f'Calculating and writing with module {module}:')
        missing_features = missing_vars.index.unique('feature')
        ds = get_features(cutout, module, missing_features, tmpdir=tmpdir)
        prepared |= set(missing_features)

        cutout.data.attrs.update(dict(prepared_features=list(prepared)))
        ds = (cutout.data.merge(ds[missing_vars.values])
              .assign_attrs(**non_bool_dict(cutout.data.attrs), **ds.attrs))

        # write data to tmp file, copy it to original data, this is much safer
        # than appending variables
        directory, filename = os.path.split(str(cutout.path))
        fd, tmp = mkstemp(suffix=filename, dir=directory)
        os.close(fd)

        with ProgressBar():
            ds.to_netcdf(tmp)

        if cutout.path.exists():
            cutout.data.close()
            cutout.path.unlink()
        os.rename(tmp, cutout.path)

        cutout.data = xr.open_dataset(cutout.path, chunks=cutout.chunks)

    return cutout



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module containing specific operations for creating cutouts from the CORDEX dataset.

DEPRECATED
----------

The cordex dataset module has not been ported to Atlite v0.2, yet. Use Atlite v0.0.2 to use it,
for the time being!
"""

import pandas as pd
import numpy as np
import xarray as xr
import pyproj
from itertools import groupby
from operator import itemgetter
import os
import glob

# Model and CRS Settings
model = 'MPI-M-MPI-ESM-LR'

crs = 4326 # TODO
# something like the following is correct
# crs = pyproj.crs.DerivedGeographicCRS(4326, pcrs.coordinate_operation.RotatedLatitudeLongitudeConversion(??))
# RotProj(dict(proj='ob_tran', o_proj='latlong', lon_0=180, o_lon_p=-162, o_lat_p=39.25))


def rename_and_clean_coords(ds):
    ds = ds.rename({'rlon': 'x', 'rlat': 'y'})
    # drop some coordinates and variables we do not use
    ds = ds.drop((set(ds.coords) | set(ds.data_vars))
                 & {'bnds', 'height', 'rotated_pole'})
    return ds


def prepare_data_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.rename({oldname: newname})
        ds = ds.sel(x=xs, y=ys)

        if newname in {'influx', 'outflux'}:
            # shift averaged data to beginning of bin
            ds = ds.assign_coords(
                time=(
                    pd.to_datetime(
                        ds.coords["time"].values) -
                    pd.Timedelta(
                        hours=1.5)))
        elif newname in {'runoff'}:
            # shift and fill 6hr average data to beginning of 3hr bins
            t = pd.to_datetime(ds.coords["time"].values)
            ds = ds.reindex(
                method='bfill',
                time=(
                    t -
                    pd.Timedelta(
                        hours=3.)).union(t))

        for m in months:
            yield (year, m), ds.sel(time="{}-{}".format(year, m))


def prepare_static_data_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.rename({oldname: newname})
        ds = ds.sel(x=xs, y=ys)

        for m in months:
            yield (year, m), ds


def prepare_weather_types_cordex(fn, year, months, oldname, newname, xs, ys):
    with xr.open_dataset(fn) as ds:
        ds = ds.rename({oldname: newname})
        for m in months:
            yield (year, m), ds.sel(time="{}-{}".format(year, m))


def prepare_meta_cordex(
        xs,
        ys,
        year,
        month,
        template,
        height_config,
        module,
        model=model):
    fn = next(glob.iglob(template.format(year=year, model=model)))
    with xr.open_dataset(fn) as ds:
        ds = rename_and_clean_coords(ds)
        ds = ds.coords.to_dataset()
        meta = ds.sel(time="{}-{}".format(year, month),
                      x=xs, y=ys).load()

    xs = ds['x'].values
    ys = ds['y'].values

    height_config = height_config.copy()
    height_tasks_func = height_config.pop('tasks_func')
    height_task, = height_tasks_func(
        xs, ys, [(year, month)], meta_attrs={}, **height_config)
    height_prepare_func = height_task.pop('prepare_func')
    _, ds = height_prepare_func(**height_task)[0]

    meta['height'] = ds['height']

    return meta


def tasks_yearly_cordex(
        xs,
        ys,
        yearmonths,
        prepare_func,
        template,
        oldname,
        newname,
        meta_attrs):
    model = meta_attrs['model']

    if not isinstance(xs, slice):
        first, second, last = xs.values[[0, 1, -1]]
        xs = slice(first - 0.1 * (second - first),
                   last + 0.1 * (second - first))
    if not isinstance(ys, slice):
        first, second, last = ys.values[[0, 1, -1]]
        ys = slice(first - 0.1 * (second - first),
                   last + 0.1 * (second - first))

    return [dict(prepare_func=prepare_func,
                 xs=xs, ys=ys, oldname=oldname, newname=newname,
                 fn=next(glob.iglob(template.format(year=year, model=model))),
                 year=year, months=list(map(itemgetter(1), yearmonths)))
            for year, yearmonths in groupby(yearmonths, itemgetter(0))]


weather_data_config = {
    'influx': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='rsds', newname='influx',
                   template=os.path.join(config.cordex_dir, '{model}', 'influx', 'rsds_*_{year}*.nc')),
    'outflux': dict(tasks_func=tasks_yearly_cordex,
                    prepare_func=prepare_data_cordex,
                    oldname='rsus', newname='outflux',
                    template=os.path.join(config.cordex_dir, '{model}', 'outflux', 'rsus_*_{year}*.nc')),
    'temperature': dict(tasks_func=tasks_yearly_cordex,
                        prepare_func=prepare_data_cordex,
                        oldname='tas', newname='temperature',
                        template=os.path.join(config.cordex_dir, '{model}', 'temperature', 'tas_*_{year}*.nc')),
    'humidity': dict(tasks_func=tasks_yearly_cordex,
                     prepare_func=prepare_data_cordex,
                     oldname='hurs', newname='humidity',
                     template=os.path.join(config.cordex_dir, '{model}', 'humidity', 'hurs_*_{year}*.nc')),
    'wnd10m': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='sfcWind', newname='wnd10m',
                   template=os.path.join(config.cordex_dir, '{model}', 'wind', 'sfcWind_*_{year}*.nc')),
    'roughness': dict(tasks_func=tasks_yearly_cordex,
                      prepare_func=prepare_static_data_cordex,
                      oldname='rlst', newname='roughness',
                      template=os.path.join(config.cordex_dir, '{model}', 'roughness', 'rlst_*.nc')),
    'runoff': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_data_cordex,
                   oldname='mrro', newname='runoff',
                   template=os.path.join(config.cordex_dir, '{model}', 'runoff', 'mrro_*_{year}*.nc')),
    'height': dict(tasks_func=tasks_yearly_cordex,
                   prepare_func=prepare_static_data_cordex,
                   oldname='orog', newname='height',
                   template=os.path.join(config.cordex_dir, '{model}', 'altitude', 'orog_*.nc')),
    'CWT': dict(tasks_func=tasks_yearly_cordex,
                prepare_func=prepare_weather_types_cordex,
                oldname='CWT', newname='CWT',
                template=os.path.join(config.cordex_dir, '{model}', 'weather_types', 'CWT_*_{year}*.nc')),
}

meta_data_config = dict(
    prepare_func=prepare_meta_cordex,
    template=os.path.join(
        config.cordex_dir,
        '{model}',
        'temperature',
        'tas_*_{year}*.nc'),
    height_config=weather_data_config['height'])



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2020 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Functions for Geographic Information System.
"""

import numpy as np
import pandas as pd
import xarray as xr
import scipy as sp
import scipy.sparse
import geopandas as gpd
import rasterio as rio
import rasterio.warp
import multiprocessing as mp

from collections import OrderedDict
from pathlib import Path
from warnings import warn
from pyproj import CRS, Transformer
from shapely.ops import transform
from rasterio.warp import reproject, transform_bounds
from rasterio.mask import mask
from rasterio.features import geometry_mask
from scipy.ndimage.morphology import binary_dilation as dilation
from numpy import isin, empty
from shapely.strtree import STRtree
from tqdm import tqdm


import logging
logger = logging.getLogger(__name__)


def get_coords(x, y, time, dx=0.25, dy=0.25, dt='h', **kwargs):
    """
    Create an cutout coordinate system on the basis of slices and step sizes.

    Parameters
    ----------
    x : slice
        Numerical slices with lower and upper bound of the x dimension.
    y : slice
        Numerical slices with lower and upper bound of the y dimension.
    time : slice
        Slice with strings with lower and upper bound of the time dimension.
    dx : float, optional
        Step size of the x coordinate. The default is 0.25.
    dy : float, optional
        Step size of the y coordinate. The default is 0.25.
    dt : str, optional
        Frequency of the time coordinate. The default is 'h'. Valid are all
        pandas offset aliases.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with x, y and time variables, representing the whole coordinate
        system.
    """
    x = slice(*sorted([x.start, x.stop]))
    y = slice(*sorted([y.start, y.stop]))

    ds = xr.Dataset({'x': np.round(np.arange(-180, 180, dx), 9),
                     'y': np.round(np.arange(-90, 90, dy), 9),
                     'time': pd.date_range(start="1979", end="now", freq=dt)})
    ds = ds.assign_coords(lon=ds.coords['x'], lat=ds.coords['y'])
    ds = ds.sel(x=x, y=y, time=time)
    return ds


def spdiag(v):
    """Create a sparse diagonal matrix from a 1-dimensional array."""
    N = len(v)
    inds = np.arange(N + 1, dtype=np.int32)
    return sp.sparse.csr_matrix((v, inds[:-1], inds), (N, N))


def reproject_shapes(shapes, crs1, crs2):
    """Project a collection of shapes from one crs to another."""
    transformer = Transformer.from_crs(crs1, crs2)

    def _reproject_shape(shape):
        return transform(transformer.transform, shape)

    if isinstance(shapes, pd.Series):
        return shapes.map(_reproject_shape)
    elif isinstance(shapes, dict):
        return OrderedDict((k, _reproject_shape(v)) for k, v in shapes.items())
    else:
        return list(map(_reproject_shape, shapes))


def reproject(shapes, p1, p2):
    """
    Project a collection of shapes from one crs to another.

    Deprecated since version 0.2.
    """
    warn("reproject has been renamed to reproject_shapes", DeprecationWarning)
    return reproject_shapes(shapes, p1, p2)


reproject.__doc__ = reproject_shapes.__doc__



def compute_indicatormatrix(orig, dest, orig_crs=4326, dest_crs=4326):
    """
    Compute the indicatormatrix.

    The indicatormatrix I[i,j] is a sparse representation of the ratio
    of the area in orig[j] lying in dest[i], where orig and dest are
    collections of polygons, i.e.

    A value of I[i,j] = 1 indicates that the shape orig[j] is fully
    contained in shape dest[j].

    Note that the polygons must be in the same crs.

    Parameters
    ----------
    orig : Collection of shapely polygons
    dest : Collection of shapely polygons

    Returns
    -------
    I : sp.sparse.lil_matrix
      Indicatormatrix
    """
    orig = orig.geometry if isinstance(orig, gpd.GeoDataFrame) else orig
    dest = dest.geometry if isinstance(dest, gpd.GeoDataFrame) else dest
    dest = reproject_shapes(dest, dest_crs, orig_crs)
    indicator = sp.sparse.lil_matrix((len(dest), len(orig)), dtype=np.float)
    tree = STRtree(orig)
    idx = dict((id(o), i) for i, o in enumerate(orig))

    for i, d in enumerate(dest):
        for o in tree.query(d):
            if o.intersects(d):
                j = idx[id(o)]
                area = d.intersection(o).area
                indicator[i, j] = area / o.area

    return indicator


class ExclusionContainer:
    """Container for exclusion objects and meta data."""

    def __init__(self, crs=3035, res=100):
        """Initialize a container for excluded areas.

        Parameters
        ----------
        crs : rasterio.CRS/proj.CRS/EPSG, optional
            Base crs of the raster collection. All rasters and geometries
            diverging from this crs will be converted to it.
            The default is 3035.
        res : float, optional
            Resolution of the base raster. All diverging rasters will be
            resampled using the gdal Resampling method 'nearest'.
            The default is 100.
        """
        self.rasters = []
        self.geometries = []
        self.crs = crs
        self.res = res

    def add_raster(self, raster, codes=None, buffer=0, invert=False, nodata=255,
                   allow_no_overlap=False, crs=None):
        """
        Register a raster to the ExclusionContainer.

        Parameters
        ----------
        raster : str/rasterio.DatasetReader
            Raster or path to raster which to exclude.
        codes : int/list/function, optional
            Codes in the raster which to exclude. Can be a callable function
            which takes the mask (np.array) as argument and performs a
            elementwise condition (must not change the shape). The default is 1.
        buffer : int, optional
            Buffer around the excluded areas in units of ExclusionContainer.crs.
            Use this to create a buffer around the excluded/included area.
            The default is 0.
        invert : bool, optional
            Whether to exclude (False) or include (True) the specified areas
            of the raster. The default is False.
        allow_no_overlap:
            Allow that a raster and a shape (for which the raster will be used as
            a mask) do not overlap. In this case an array with only `nodata` is
            returned.
        crs : rasterio.CRS/EPSG
            CRS of the raster. Specify this if the raster has invalid crs.
        """
        d = dict(raster=raster, codes=codes, buffer=buffer, invert=invert,
                 nodata=nodata, allow_no_overlap=allow_no_overlap, crs=crs)
        self.rasters.append(d)

    def add_geometry(self, geometry, buffer=0, invert=False):
        """
        Register a collection of geometries to the ExclusionContainer.

        Parameters
        ----------
        geometry : str/path/geopandas.GeoDataFrame
            Path to geometries or geometries which to exclude.
        buffer : float, optional
            Buffer around the excluded areas in units of ExclusionContainer.crs.
            The default is 0.
        invert : bool, optional
            Whether to exclude (False) or include (True) the specified areas
            of the geometries. The default is False.
        """
        d = dict(geometry=geometry, buffer=buffer, invert=invert)
        self.geometries.append(d)


    def open_files(self):
        """Open rasters and load geometries."""
        for d in self.rasters:
            raster = d['raster']
            if isinstance(raster, (str, Path)):
                raster = rio.open(raster)
            else:
                assert isinstance(raster, rio.DatasetReader)
            if (not raster.crs.is_valid if raster.crs is not None else True):
                if d['crs']:
                    raster._crs = CRS(d['crs'])
                else:
                    raise ValueError(f'CRS of {raster} is invalid, please '
                                     'provide it.')
            d['raster'] = raster

        for d in self.geometries:
            geometry = d['geometry']
            if isinstance(geometry, (str, Path)):
                geometry = gpd.read_file(geometry)
            if isinstance(geometry, gpd.GeoDataFrame):
                geometry = geometry.geometry
            assert isinstance(geometry, gpd.GeoSeries)
            assert geometry.crs is not None
            geometry = geometry.to_crs(self.crs)
            if d.get('buffer', 0) and not d.get('_buffered', False):
                geometry = geometry.buffer(d['buffer'])
                d['_buffered'] = True
            d['geometry'] = geometry

    @property
    def all_closed(self):
        """Check whether all files in the raster container are closed."""
        return all(isinstance(d['raster'], (str, Path)) for d in self.rasters)

    @property
    def all_open(self):
        """Check whether all files in the raster container are open."""
        return all(isinstance(d['raster'], rio.DatasetReader) for d in self.rasters)

    def __repr__(self):
        return (f"Exclusion Container"
                f"\n registered rasters: {len(self.rasters)} "
                f"\n registered geometry collections: {len(self.geometries)}"
                f"\n CRS: {self.crs} - Resolution: {self.res}")


def padded_transform_and_shape(bounds, res):
    """
    Get the (transform, shape) tuple of a raster with resolution `res` and
    bounds `bounds`.
    """
    left, bottom = [(b // res)* res for b in bounds[:2]]
    right, top = [(b // res + 1) * res for b in bounds[2:]]
    shape = int((top - bottom) / res), int((right - left) / res)
    return rio.Affine(res, 0, left, 0, -res, top), shape


def projected_mask(raster, geom, transform=None, shape=None, crs=None,
                   allow_no_overlap=False, **kwargs):
    """Load a mask and optionally project it to target resolution and shape."""
    nodata = kwargs.get('nodata', 255)
    kwargs.setdefault('indexes', 1)
    if geom.crs != raster.crs:
        geom = geom.to_crs(raster.crs)

    if allow_no_overlap:
        try:
            masked, transform_ = mask(raster, geom, crop=True, **kwargs)
        except ValueError:
            res = raster.res[0]
            transform_, shape = padded_transform_and_shape(geom.total_bounds, res)
            masked = np.full(shape, nodata)
    else:
        masked, transform_ = mask(raster, geom, crop=True, **kwargs)

    if transform is None or (transform_ == transform and shape == masked.shape):
        return masked, transform_

    assert shape is not None and crs is not None
    return rio.warp.reproject(masked, empty(shape), src_crs=raster.crs,
                              dst_crs=crs, src_transform=transform_,
                              dst_transform=transform, dst_nodata=nodata)


def pad_extent(src, src_transform, dst_transform, src_crs, dst_crs, **kwargs):
    """
    Pad the extent of `src` by an equivalent of one cell of the target raster.

    This ensures that the array is large enough to not be treated as nodata in
    all cells of the destination raster. If src.ndim > 2, the function expects
    the last two dimensions to be y,x.
    Additional keyword arguments are used in `np.pad()`.
    """
    if src.size == 0:
        return src, src_transform

    left, top, right, bottom = *(src_transform*(0,0)), *(src_transform*(1,1))
    covered = transform_bounds(src_crs, dst_crs, left, bottom, right, top)
    covered_res = min(abs(covered[2] - covered[0]), abs(covered[3] - covered[1]))
    pad = int(dst_transform[0] // covered_res * 1.1)

    kwargs.setdefault('mode', 'constant')

    if src.ndim == 2:
        return rio.pad(src, src_transform, pad, **kwargs)

    npad = ((0,0),) * (src.ndim - 2) + ((pad, pad), (pad, pad))
    padded = np.pad(src, npad, **kwargs)
    transform = list(src_transform)
    transform[2] -= pad * transform[0]
    transform[5] -= pad * transform[4]
    return padded, rio.Affine(*transform[:6])


def shape_availability(geometry, excluder):
    """
    Compute the eligible area in one or more geometries.

    Parameters
    ----------
    geometry : geopandas.Series
        Geometry of which the eligible area is computed. If the series contains
        more than one geometry, the eligble area of the combined geometries is
        computed.
    excluder : atlite.gis.ExclusionContainer
        Container of all meta data or objects which to exclude, i.e.
        rasters and geometries.

    Returns
    -------
    masked : np.array
        Mask whith eligible raster cells indicated by 1 and excluded cells by 0.
    transform : rasterion.Affine
        Affine transform of the mask.

    """
    exclusions = []
    if not excluder.all_open:
        excluder.open_files()
    assert geometry.crs == excluder.crs

    bounds = rio.features.bounds(geometry)
    transform, shape = padded_transform_and_shape(bounds, res=excluder.res)
    masked = geometry_mask(geometry, shape, transform).astype(int)
    exclusions.append(masked)

    # For the following: 0 is eligible, 1 in excluded
    raster = None
    for d in excluder.rasters:
        # allow reusing preloaded raster with different post-processing
        if raster != d['raster']:
            raster = d['raster']
            kwargs_keys = ['allow_no_overlap', 'nodata']
            kwargs = {k: v for k, v in d.items() if k in kwargs_keys}
            masked, transform = projected_mask(d['raster'], geometry, transform,
                                               shape, excluder.crs, **kwargs)
        if d['codes']:
            if callable(d['codes']):
                masked_ = d['codes'](masked)
            else:
                masked_ = isin(masked, d['codes'])
        else:
            masked_ = masked

        if d['invert']:
            masked_ = ~(masked_).astype(bool)
        if d['buffer']:
            iterations = int(d['buffer'] / excluder.res) + 1
            masked_ = dilation(masked_, iterations=iterations).astype(int)

        exclusions.append(masked_.astype(int))

    for d in excluder.geometries:
        masked = ~geometry_mask(d['geometry'], shape, transform,
                                invert=d['invert'])
        exclusions.append(masked.astype(int))

    return (sum(exclusions) == 0).astype(float), transform


def shape_availability_reprojected(geometry, excluder, dst_transform, dst_crs,
                                   dst_shape):
    """
    Compute and reproject the eligible area of one or more geometries.

    The function executes `shape_availability` and reprojects the calculated
    mask onto a new raster defined by (dst_transform, dst_crs, dst_shape).
    Before reprojecting, the function pads the mask such all non-nodata data
    points  are projected in full cells of the target raster. The ensures that
    all data within the mask are projected correclty (GDAL inherent 'problem').

    ----------
    geometry : geopandas.Series
        Geometry in which the eligible area is computed. If the series contains
        more than one geometry, the eligble area of the combined geometries is
        computed.
    excluder : atlite.gis.ExclusionContainer
        Container of all meta data or objects which to exclude, i.e.
        rasters and geometries.
    dst_transform : rasterio.Affine
        Transform of the target raster.
    dst_crs : rasterio.CRS/proj.CRS
        CRS of the target raster.
    dst_shape : tuple
        Shape of the target raster.

    masked : np.array
        Average share of available area per grid cell. 0 indicates excluded,
        1 is fully included.
    transform : rasterio.Affine
        Affine transform of the mask.

    """
    masked, transform = shape_availability(geometry, excluder)
    masked, transform = pad_extent(masked, transform, dst_transform,
                                   excluder.crs, dst_crs)
    return rio.warp.reproject(masked, empty(dst_shape), resampling=5,
                              src_transform=transform,
                              dst_transform=dst_transform,
                              src_crs=excluder.crs, dst_crs=dst_crs,)


def _init_process(shapes_, excluder_, dst_transform_, dst_crs_, dst_shapes_):
    global shapes, excluder, dst_transform, dst_crs, dst_shapes
    shapes, excluder = shapes_, excluder_
    dst_transform, dst_crs, dst_shapes = dst_transform_, dst_crs_, dst_shapes_


def _process_func(i):
    args = (excluder, dst_transform, dst_crs, dst_shapes)
    return shape_availability_reprojected(shapes.loc[[i]], *args)[0]


def compute_availabilitymatrix(cutout, shapes, excluder, nprocesses=None,
                               disable_progressbar=False):
    """
    Compute the eligible share within cutout cells in the overlap with shapes.

    For parallel calculation (nprocesses not None) the excluder must not be
    initialized and all raster references must be strings. Otherwise processes
    are colliding when reading from one common rasterio.DatasetReader.

    Parameters
    ----------
    cutout : atlite.Cutout
        Cutout which the availability matrix is aligned to.
    shapes : geopandas.Series/geopandas.DataFrame
        Geometries for which the availabilities are calculated.
    excluder : atlite.gis.ExclusionContainer
        Container of all meta data or objects which to exclude, i.e.
        rasters and geometries.
    nprocesses : int, optional
        Number of processes to use for calculating the matrix. The paralle-
        lization can heavily boost the calculation speed. The default is None.
    disable_progressbar: bool, optional
        Disable the progressbar if nprocesses is not None. Then the `map`
        function instead of the `imap` function is used for the multiprocessing
        pool. This speeds up the calculation.

    Returns
    -------
    availabilities : xr.DataArray
        DataArray of shape (|shapes|, |y|, |x|) containing all the eligible
        share of cutout cell (x,y) in the overlap with shape i.

    Notes
    -----
    The rasterio (or GDAL) average downsampling returns different results
    dependent on how the target raster (the cutout raster) is spanned.
    Either it is spanned from the top left going downwards,
    e.g. Affine(0.25, 0, 0, 0, -0.25, 50), or starting in the
    lower left corner and going up, e.g. Affine(0.25, 0, 0, 0, 0.25, 50).
    Here we stick to the top down version which is why we use
    `cutout.transform_r` and flipping the y-axis in the end.

    """
    availability = []
    shapes = shapes.geometry if isinstance(shapes, gpd.GeoDataFrame) else shapes
    shapes = shapes.to_crs(excluder.crs)


    args = (excluder, cutout.transform_r, cutout.crs, cutout.shape)
    tqdm_kwargs = dict(ascii=False, unit=' gridcells', total=len(shapes),
                       desc='Compute availability matrix')
    if nprocesses is None:
        for i in tqdm(shapes.index, **tqdm_kwargs):
            _ = shape_availability_reprojected(shapes.loc[[i]], *args)[0]
            availability.append(_)
    else:
        assert excluder.all_closed, ('For parallelization all raster files '
                                     'in excluder must be closed')
        kwargs = {'initializer': _init_process,
                   'initargs': (shapes, *args),
                   'maxtasksperchild': 20,
                   'processes': nprocesses}
        with mp.Pool(**kwargs) as pool:
            if disable_progressbar:
                availability = list(pool.map(_process_func, shapes.index))
            else:
                availability = list(tqdm(pool.imap(_process_func, shapes.index),
                                    **tqdm_kwargs))



    availability = np.stack(availability)[:, ::-1] # flip axis, see Notes
    coords=[(shapes.index), ('y', cutout.data.y), ('x', cutout.data.x)]
    return xr.DataArray(availability, coords=coords)



def maybe_swap_spatial_dims(ds, namex='x', namey='y'):
    """Swap order of spatial dimensions according to atlite concention."""
    swaps = {}
    lx, rx = ds.indexes[namex][[0, -1]]
    ly, uy = ds.indexes[namey][[0, -1]]

    if lx > rx:
        swaps[namex] = slice(None, None, -1)
    if uy < ly:
        swaps[namey] = slice(None, None, -1)

    return ds.isel(**swaps) if swaps else ds


def _as_transform(x, y):
    lx, rx = x[[0, -1]]
    ly, uy = y[[0, -1]]

    dx = float(rx - lx) / float(len(x) - 1)
    dy = float(uy - ly) / float(len(y) - 1)

    return rio.Affine(dx, 0, lx - dx/2, 0, dy, ly - dy/2)


def regrid(ds, dimx, dimy, **kwargs):
    """
    Interpolate Dataset or DataArray `ds` to a new grid, using rasterio's
    reproject facility.

    See also: https://mapbox.github.io/rasterio/topics/resampling.html

    Parameters
    ----------
    ds : xr.Dataset|xr.DataArray
      N-dim data on a spatial grid
    dimx : pd.Index
      New x-coordinates in destination crs.
      dimx.name MUST refer to x-coord of ds.
    dimy : pd.Index
      New y-coordinates in destination crs.
      dimy.name MUST refer to y-coord of ds.
    **kwargs :
      Arguments passed to rio.wrap.reproject; of note:
      - resampling is one of gis.Resampling.{average,cubic,bilinear,nearest}
      - src_crs, dst_crs define the different crs (default: EPSG 4326, ie latlong)
    """
    namex = dimx.name
    namey = dimy.name

    ds = maybe_swap_spatial_dims(ds, namex, namey)

    src_transform = _as_transform(ds.indexes[namex], ds.indexes[namey])
    dst_transform = _as_transform(dimx, dimy)
    dst_shape = len(dimy), len(dimx)

    kwargs.update(dst_transform=dst_transform)
    kwargs.setdefault("src_crs", CRS.from_epsg(4326))
    kwargs.setdefault("dst_crs", CRS.from_epsg(4326))

    def _reproject(src, **kwargs):
        shape = src.shape[:-2] + dst_shape
        src, trans = pad_extent(src, src_transform, dst_transform,
                                kwargs['src_crs'], kwargs['dst_crs'], mode='edge')

        return rio.warp.reproject(src, empty(shape), src_transform=trans, **kwargs)[0]


    data_vars = ds.data_vars.values() if isinstance(ds, xr.Dataset) else (ds,)
    dtypes = {da.dtype for da in data_vars}
    assert len(dtypes) == 1, \
        "regrid can only reproject datasets with homogeneous dtype"

    return (xr.apply_ufunc(_reproject,
                           ds,
                           input_core_dims=[[namey, namex]],
                           output_core_dims=[['yout', 'xout']],
                           output_dtypes=[dtypes.pop()],
                           dask_gufunc_kwargs =
                               dict(output_sizes={'yout': dst_shape[0],
                                                  'xout': dst_shape[1]}),
                           dask='parallelized',
                           kwargs=kwargs)
            .rename({'yout': namey, 'xout': namex})
            .assign_coords(**{namey: (namey, dimy, ds.coords[namey].attrs),
                              namex: (namex, dimx, ds.coords[namex].attrs)})
            .assign_attrs(**ds.attrs))



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Functions for use in conjunction with wind data generation.
"""

import numpy as np

import logging
logger = logging.getLogger(__name__)


def extrapolate_wind_speed(ds, to_height, from_height=None):
    """Extrapolate the wind speed from a given height above ground to another.

    If ds already contains a key refering to wind speeds at the desired to_height,
    no conversion is done and the wind speeds are directly returned.

    Extrapolation of the wind speed follows the logarithmic law as desribed in [1].

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the wind speed time-series at 'from_height' with key
        'wnd{height:d}m' and the surface orography with key 'roughness' at the
        geographic locations of the wind speeds.
    from_height : int
        (Optional)
        Height (m) from which the wind speeds are interpolated to 'to_height'.
        If not provided, the closest height to 'to_height' is selected.
    to_height : int|float
        Height (m) to which the wind speeds are extrapolated to.

    Returns
    -------
    da : xarray.DataArray
        DataArray containing the extrapolated wind speeds. Name of the DataArray
        is 'wnd{to_height:d}'.

    References
    ----------
    [1] Equation (2) in Andresen, G. et al (2015): 'Validation of Danish wind
    time series from a new global renewable energy atlas for energy system
    analysis'.

    [2] https://en.wikipedia.org/w/index.php?title=Roughness_length&oldid=862127433,
    Retrieved 2019-02-15.
    """

    # Fast lane
    to_name = "wnd{h:0d}m".format(h=int(to_height))
    if to_name in ds:
        return ds[to_name]

    if from_height is None:
        # Determine closest height to to_name
        heights = np.asarray([int(s[3:-1]) for s in ds if s.startswith("wnd")])

        if len(heights) == 0:
            raise AssertionError("Wind speed is not in dataset")

        from_height = heights[np.argmin(np.abs(heights - to_height))]

    from_name = "wnd{h:0d}m".format(h=int(from_height))

    # Wind speed extrapolation
    wnd_spd = ds[from_name] * (np.log(to_height / ds['roughness'])
                               / np.log(from_height / ds['roughness']))

    wnd_spd.attrs.update({"long name":
                          "extrapolated {ht} m wind speed using logarithmic "
                          "method with roughness and {hf} m wind speed"
                          "".format(ht=to_height, hf=from_height),
                          "units": "m s**-1"})

    return wnd_spd.rename(to_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  6 15:23:13 2020

@author: fabian
"""

# IDEAS for tests

import pytest
import geopandas as gpd
import xarray as xr
import numpy as np
import rasterio as rio
import rasterio.warp
from atlite import Cutout
from atlite.gis import ExclusionContainer, shape_availability, pad_extent, regrid
from shapely.geometry import box
from atlite.gis import padded_transform_and_shape
from numpy import isclose, allclose
from xarray.testing import assert_allclose, assert_equal
import pandas as pd



TIME = '2013-01-01'
X0 = -4.
Y0 = 56.
X1 = 1.5
Y1 = 61.
raster_clip = 0.25 # this rastio is excluded (True) in the raster


@pytest.fixture
def ref():
    return Cutout(path="creation_ref", module="era5", bounds=(X0, Y0, X1, Y1), time=TIME)

@pytest.fixture(scope='session')
def raster(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = (X0, Y0, X1, Y1) # same as in test_gis.py
    res = 0.01
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = np.random.rand(*shape) < raster_clip
    mask = mask.astype(rio.int32)
    path = tmp_path / 'raster.tif'
    with rio.open(path, 'w', driver='GTiff', transform=transform, crs=4326,
                  width=shape[1], height=shape[0], count=1, dtype=mask.dtype) as dst:
        dst.write(mask, indexes=1)
    return path

@pytest.fixture(scope='session')
def raster_reproject(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = rio.warp.transform_bounds(4326, 3035, X0, Y0, X1, Y1)
    res = 1000
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = np.random.rand(*shape) < raster_clip
    mask = mask.astype(rio.int32)
    path = tmp_path / 'raster_reproject.tif'
    with rio.open(path, 'w', driver='GTiff', transform=transform, crs=3035,
                  width=shape[1], height=shape[0], count=1, dtype=mask.dtype) as dst:
        dst.write(mask, indexes=1)
    return path

@pytest.fixture(scope='session')
def raster_codes(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = (X0, Y0, X1, Y1) # same as in test_gis.py
    res = 0.01
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = (np.random.rand(*shape) * 100).astype(int)
    mask = mask.astype(rio.int32)
    path = tmp_path / 'raster_codes.tif'
    with rio.open(path, 'w', driver='GTiff', transform=transform, crs=4326,
                  width=shape[1], height=shape[0], count=1, dtype=mask.dtype) as dst:
        dst.write(mask, indexes=1)
    return path


def test_transform():
    """Test the affine transform. It always has to point to cell origin."""
    cutout = Cutout(path="resolution", module="era5",
                    time=slice('2013-01-01', '2013-01-01'),
                    x=slice(X0, X1), y = slice(Y0, Y1))
    assert cutout.transform * (0.5, 0.5) == (X0, Y0)
    assert cutout.transform * cutout.shape[::-1] == (X1 + cutout.dx/2,
                                                     Y1 + cutout.dy/2)



def test_grid_coords(ref):
    gcoords = ref.grid[['x', 'y']]
    spatial = ref.data.stack(spatial=['y', 'x'])['spatial'].data
    spatial = np.array([[s[1], s[0]] for s in spatial])
    np.testing.assert_equal(gcoords, spatial)


def test_extent(ref):
    pad = 0.25/2
    np.testing.assert_array_equal(ref.extent,[X0-pad, X1+pad, Y0-pad, Y1+pad])

# Note that bounds is the same as extent but in different order.
def test_bounds(ref):
    np.testing.assert_array_equal(ref.bounds, ref.grid.total_bounds)


def test_regrid():
    """Test the atlite.gis.regrid function with average resampling."""
    # define blocks
    A = 0.25
    B = 0.5
    C = 0.3
    D = 0.1
    ones = np.ones((4,4))
    fine = np.block([[ones*A, ones*B], [ones*C, ones*D]])
    # add coordinates
    finecoords = np.arange(0.5, 8, 1)
    fine = xr.DataArray(fine, coords=[('y', finecoords), ('x', finecoords)])

    coarsecoords = np.arange(2, 8, 4)
    coarse = xr.DataArray(np.nan, coords=[('y', coarsecoords), ('x', coarsecoords)])

    # apply average resampling
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.array([[A, B], [C,D]])
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()

    # now test multiple layers

    fine = xr.concat([fine]*10, pd.Index(range(10), name='z'))
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.stack([np.array([[A, B], [C,D]])]*10)
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()

    # now let the target grid cover a subarea of the original
    fine = fine.sel(z=0, drop=True)
    coarsecoords = np.arange(1, 6, 2)
    coarse = xr.DataArray(np.nan, coords=[('y', coarsecoords), ('x', coarsecoords)])

    # apply average resampling
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.array([[A, A, B], [A, A, B], [C, C, D]])
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()




def test_pad_extent():
    """Test whether padding works with arrays of dimension > 2 """
    src = np.ones((3, 2))
    src_trans = rio.Affine(1, 0, 0, 0, 1, 0)
    dst_trans = rio.Affine(2, 0, 0, 0, 2, 0)
    crs = 4326

    padded, trans = pad_extent(src, src_trans, dst_trans, crs, crs)
    src = np.ones((1, 3, 2))
    padded_ndim, trans_ndim = pad_extent(src, src_trans, dst_trans, crs, crs)
    assert (padded_ndim[0] == padded).all()
    assert trans == trans_ndim

    #second check with large shape
    src = np.ones((1, 2, 3, 2))
    padded_ndim, trans_ndim = pad_extent(src, src_trans, dst_trans, crs, crs)
    assert (padded_ndim[0, 0] == padded).all()
    assert trans == trans_ndim

    #other way round, here it should not pad, since target resolution is lower
    padded_r, trans_r = pad_extent(src, dst_trans, src_trans, crs, crs)
    assert (padded_r == src).all()
    assert trans_r == dst_trans



def test_indicator_matrix(ref):
    # This should be the grid cell at the lower left corner
    cell = ref.grid.geometry[0]
    indicator = ref.indicatormatrix([cell])
    assert indicator[0, 0] == 1.
    assert indicator.sum() == 1
    # This should be the grid cell at the lower left corner
    cell = ref.grid.geometry.iloc[-2]
    indicator = ref.indicatormatrix([cell])
    assert indicator[0, -2] == 1.
    assert indicator.sum() == 1


def test_availability_matrix_flat(ref):
    """
    Indicator matrix and availability matrix for an empty excluder must be
    the same.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y1-1)], crs=ref.crs).rename_axis('shape')
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords['y'], ref.coords['x']])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder)
    assert np.allclose(I, ds.sum('shape'))


def test_availability_matrix_flat_parallel(ref):
    """
    Same as `test_availability_matrix_flat` but parallel and without progressbar.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y1-1)], crs=ref.crs).rename_axis('shape')
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords['y'], ref.coords['x']])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder, nprocesses=2)
    assert np.allclose(I, ds.sum('shape'))


def test_availability_matrix_flat_wo_progressbar(ref):
    """
    Same as `test_availability_matrix_flat` but without progressbar.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y1-1)], crs=ref.crs).rename_axis('shape')
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords['y'], ref.coords['x']])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder, disable_progressbar=True)
    assert np.allclose(I, ds.sum('shape'))


def test_availability_matrix_flat_parallel_wo_progressbar(ref):
    """
    Same as `test_availability_matrix_flat` but parallel and without progressbar.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y1-1)], crs=ref.crs).rename_axis('shape')
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords['y'], ref.coords['x']])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder, nprocesses=2, disable_progressbar=True)
    assert np.allclose(I, ds.sum('shape'))



def test_shape_availability_area(ref):
    """Area of the mask and the shape must be close."""
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y1-1)], crs=ref.crs)
    res = 100
    excluder = ExclusionContainer(res=res)

    with pytest.raises(AssertionError):
        masked, transform = shape_availability(shapes, excluder)

    shapes = shapes.to_crs(3035)
    masked, transform = shape_availability(shapes, excluder)
    assert np.isclose(shapes.area, masked.sum() * res ** 2)



def test_exclusioncontainer_geometries():
    crs = 3035
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=crs)
    exclude = gpd.GeoSeries([box(X0/2+X1/2, Y0/2+Y1/2, X1, Y1)], crs=crs)
    res = 0.01

    excluder = ExclusionContainer(crs, res=res)
    excluder.add_geometry(exclude, buffer=1)
    excluder.open_files()
    assert (excluder.geometries[0]['geometry'] != exclude).all()
    excluder.open_files()
    buffered = excluder.geometries[0]['geometry']
    # open again and check that the buffer remains the same
    assert (excluder.geometries[0]['geometry'] == buffered).all()

    # should take GeoDataFrames and the result is the same
    excluder = ExclusionContainer(crs, res=res)
    excluder.add_geometry(exclude.to_frame('geometry'), buffer=1)
    excluder.open_files()
    assert (excluder.geometries[0]['geometry'] == buffered).all()



def test_shape_availability_exclude_geometry(ref):
    """
    When excluding the quarter of the geometry, the eligible area must be a
    forth. Test the inverted case too.
    """
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    exclude = gpd.GeoSeries([box(X0/2+X1/2, Y0/2+Y1/2, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_geometry(exclude)
    masked, transform = shape_availability(shapes, excluder)
    area = shapes.geometry[0].area # get area without warning
    assert isclose(3*area/4, masked.sum() * res ** 2)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_geometry(exclude, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    area = shapes.geometry[0].area # get area without warning
    assert isclose(area/4, masked.sum() * res ** 2)



def test_shape_availability_exclude_raster(ref, raster):
    """When excluding the half of the geometry, the eligible area must be half."""
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 2) == (1 - raster_clip)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 2) == raster_clip

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, buffer=res)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    # should be close to zero
    assert round(ratio, 2) < (1-raster_clip)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, buffer=res, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio2 = masked.sum() / masked.size
    # for the case that we have more excluded area and this is buffered
    if raster_clip < 0.5:
        assert ratio >= ratio2
    else:
        assert ratio <= ratio2


def test_shape_availability_excluder_partial_overlap(ref, raster):
    """Test behavior, when a raster only overlaps half of the geometry."""
    bounds = X0-2, Y0, X0+2, Y1
    area = abs((bounds[2] - bounds[0]) * (bounds[3] - bounds[1]))
    shapes = gpd.GeoSeries([box(*bounds)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, codes=[0, 1])
    masked, transform = shape_availability(shapes, excluder)
    assert (masked.sum() * (res ** 2) == area/2)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, nodata=0)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked.sum() * (res ** 2) > area/2)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, nodata=1)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked.sum() * (res ** 2) < area/2)



def test_shape_availability_excluder_raster_no_overlap(ref, raster):
    """Check if the allow_no_overlap flag works."""
    bounds = X0-10., Y0-10., X0-2., Y0-2.
    area = abs((bounds[2] - bounds[0]) * (bounds[3] - bounds[1]))
    shapes = gpd.GeoSeries([box(*bounds)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster)
    with pytest.raises(ValueError):
        masked, transform = shape_availability(shapes, excluder)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked == 0).all()

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True, codes=[1, 255], invert=True)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked.sum() * (res ** 2) == area)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True, nodata=0)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked.sum() * (res ** 2) == area)



def test_availability_matrix_rastered(ref, raster):
    """
    Availability matrix with a non-zero raster must have less available area
    than the Indicator matrix.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y0/2 + Y1/2),
                            box(X0+1, Y0/2 + Y1/2, X1-1, Y1-1)],
                           crs=ref.crs).rename_axis('shape')
    I = np.asarray(ref.indicatormatrix(shapes).todense())
    I = I.reshape(shapes.shape + ref.shape)
    I = xr.DataArray(I, coords=[('shape', shapes.index), ref.coords['y'],
                                ref.coords['x']])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    excluder.add_raster(raster)
    ds = ref.availabilitymatrix(shapes, excluder)
    eligible_share = 1 - raster_clip

    assert isclose(I.sum() * eligible_share, ds.sum(), atol=5)
    assert_allclose(I.sum(['x', 'y']) * eligible_share, ds.sum(['x', 'y']), atol=5)

    excluder = ExclusionContainer(ref.crs, res=0.01)
    excluder.add_raster(raster)
    assert_equal(ds, ref.availabilitymatrix(shapes, excluder, 2))


def test_availability_matrix_rastered_repro(ref, raster_reproject):
    """
    Availability matrix with a non-zero raster must have less available area
    than the Indicator matrix. Test this with a raster of a different crs.
    """
    shapes = gpd.GeoSeries([box(X0+1, Y0+1, X1-1, Y0/2 + Y1/2),
                            box(X0+1, Y0/2 + Y1/2, X1-1, Y1-1)],
                            crs=ref.crs).rename_axis('shape')
    I = np.asarray(ref.indicatormatrix(shapes).todense())
    I = I.reshape(shapes.shape + ref.shape)
    I = xr.DataArray(I, coords=[('shape', shapes.index), ref.coords['y'],
                                ref.coords['x']])
    excluder = ExclusionContainer()
    excluder.add_raster(raster_reproject)
    ds = ref.availabilitymatrix(shapes, excluder)
    eligible_share = 1 - raster_clip

    assert isclose(I.sum() * eligible_share, ds.sum(), atol=5)
    assert_allclose(I.sum(['x', 'y']) * eligible_share, ds.sum(['x', 'y']), atol=5)



def test_shape_availability_exclude_raster_codes(ref, raster_codes):
    """Test exclusion of multiple raster codes."""
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=range(20))
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 1) == 0.8

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=range(20), invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 1) == 0.2

    # test with a function
    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=lambda x: x<20, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    assert ratio == masked.sum() / masked.size




# -*- coding: utf-8 -*-
"""
Spyder Editor

This is a temporary script file.
"""

import pandas as pd

import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection
import matplotlib.pyplot as plt

fig, ax = plt.subplots(figsize = (12, 5))
ax.axis('off')

linebreak = '\n    '

climatedata = ['Temperature', 'Downward short-wave radiation',
               'Upward short-wave radiation', 'Wind', 'Runoff',
               'Surface roughness', 'Height maps', 'Soil temperature']
processeddata = [f'Wind power generation {linebreak}for a given turbine type',
                 f'Solar PV power generation {linebreak}for a given panel type',
                 'Solar thermal collector heat output',
                 'Hydroelectric inflow (simplified)',
                 f'Heating demand {linebreak}'
                 '(based on the degree-day approximation)']

climatestr = '\n' + '\n\n'.join([' ◦ ' + s for s in climatedata]) + '\n'

processedstr = '\n' + '\n\n\n'.join([' ◦ ' + s for s in processeddata]) + '\n'

# defaults for boxes and arrows
kwargs = dict(verticalalignment='center', fontsize=14, color='#545454',)
arrowkwargs = dict(head_width=.2, width=0.13, head_length=0.05, edgecolor='white',
                   length_includes_head=True, color='lightgray', alpha=1)
y = 0.5

# First arrow
ax.text(0.01, y, ' Retrieve Data', fontsize=14, color='gray',
        verticalalignment='center')
ax.arrow(0.01, y, 0.14, 0., **arrowkwargs)

# First box
ax.text(0.17, y, climatestr, **kwargs,
        bbox=dict(facecolor='indianred', alpha=0.5, edgecolor='None',
                  boxstyle='round'))

# Second arrow
ax.text(0.5, y, ' Process Data', fontsize=14, color='gray', verticalalignment='center')
ax.arrow(0.5, y, 0.14, 0, **arrowkwargs)

# Second Box
ax.text(0.66, y, processedstr, **kwargs,
        bbox=dict(facecolor='olivedrab', alpha=0.5, edgecolor='None',
                  boxstyle='round'))


fig.tight_layout(pad=0)
fig.savefig('workflow.png', dpi=150)



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module for providing access to external ressources, like windturbine or pv panel configurations.
"""

from .utils import arrowdict
import yaml
from operator import itemgetter
import numpy as np
from scipy.signal import fftconvolve
from pathlib import Path
import requests
import pandas as pd
import json
import re
import pkg_resources

import logging
logger = logging.getLogger(name=__name__)


RESOURCE_DIRECTORY = Path(
    pkg_resources.resource_filename(
        __name__, "resources"))
WINDTURBINE_DIRECTORY = RESOURCE_DIRECTORY / "windturbine"
SOLARPANEL_DIRECTORY = RESOURCE_DIRECTORY / "solarpanel"


def get_windturbineconfig(turbine):
    """Load the wind 'turbine' configuration.

    The configuration can either be one from local storage, then 'turbine' is
    considered part of the file base name '<turbine>.yaml' in config.windturbine_dir.
    Alternatively the configuration can be downloaded from the Open Energy Database (OEDB),
    in which case 'turbine' is a dictionary used for selecting a turbine from the database.

    Parameters
    ----------
    turbine : str
        Name of the local turbine file.
        Alternatively a dict for selecting a turbine from the Open Energy
        Database, in this case the key 'source' should be contained. For all
        other key arguments to retrieve the matching turbine, see
        atlite.resource.download_windturbineconfig() for details.
    """

    if isinstance(turbine, str) and turbine.startswith("oedb:"):
        return get_oedb_windturbineconfig(turbine[len("oedb:"):])

    if isinstance(turbine, str):
        if not turbine.endswith(".yaml"):
            turbine += ".yaml"

        turbine = WINDTURBINE_DIRECTORY / turbine

    with open(turbine, "r") as f:
        conf = yaml.safe_load(f)

    return dict(
        V=np.array(conf['V']),
        POW=np.array(conf['POW']),
        hub_height=conf['HUB_HEIGHT'],
        P=np.max(conf['POW'])
    )


def get_solarpanelconfig(panel):
    """Load the 'panel'.yaml file from local disk and provide a solar panel dict."""

    if isinstance(panel, str):
        if not panel.endswith(".yaml"):
            panel += ".yaml"

        panel = SOLARPANEL_DIRECTORY / panel

    with open(panel, "r") as f:
        conf = yaml.safe_load(f)

    return conf


def solarpanel_rated_capacity_per_unit(panel):
    # unit is m^2 here

    if isinstance(panel, (str, Path)):
        panel = get_solarpanelconfig(panel)

    model = panel.get('model', 'huld')
    if model == 'huld':
        return panel['efficiency']
    elif model == 'bofinger':
        # one unit in the capacity layout is interpreted as one panel of a
        # capacity (A + 1000 * B + log(1000) * C) * 1000W/m^2 * (k / 1000)
        A, B, C = itemgetter('A', 'B', 'C')(panel)
        return (A + B * 1000. + C * np.log(1000.)) * 1e3


def windturbine_rated_capacity_per_unit(turbine):
    if isinstance(turbine, (str, Path)):
        turbine = get_windturbineconfig(turbine)

    return turbine['P']


def windturbine_smooth(turbine, params=None):
    '''
    Smooth the powercurve in `turbine` with a gaussian kernel

    Parameters
    ----------
    turbine : dict
        Turbine config with at least V and POW
    params : dict
        Allows adjusting fleet availability eta, mean Delta_v and
        stdev sigma. Defaults to values from Andresen's paper: 0.95,
        1.27 and 2.29, respectively.

    Returns
    -------
    turbine : dict
        Turbine config with a smoothed power curve

    References
    ----------
    G. B. Andresen, A. A. Søndergaard, M. Greiner, Validation of
    Danish wind time series from a new global renewable energy atlas
    for energy system analysis, Energy 93, Part 1 (2015) 1074–1088.
    '''

    if params is None:
        params = {}

    eta = params.get('eta', 0.95)
    Delta_v = params.get('Delta_v', 1.27)
    sigma = params.get('sigma', 2.29)

    def kernel(v_0):
        # all velocities in m/s
        return (1. / np.sqrt(2 * np.pi * sigma * sigma) *
                np.exp(-(v_0 - Delta_v) * (v_0 - Delta_v) / (2 * sigma * sigma)))

    def smooth(velocities, power):
        # interpolate kernel and power curve to the same, regular velocity grid
        velocities_reg = np.linspace(-50., 50., 1001)
        power_reg = np.interp(velocities_reg, velocities, power)
        kernel_reg = kernel(velocities_reg)

        # convolve power and kernel
        # the downscaling is necessary because scipy expects the velocity
        # increments to be 1., but here, they are 0.1
        convolution = 0.1 * fftconvolve(power_reg, kernel_reg, mode='same')

        # sample down so power curve doesn't get too long
        velocities_new = np.linspace(0., 35., 72)
        power_new = eta * np.interp(velocities_new,
                                    velocities_reg, convolution)

        return velocities_new, power_new

    turbine = turbine.copy()
    turbine['V'], turbine['POW'] = smooth(turbine['V'], turbine['POW'])
    turbine['P'] = np.max(turbine['POW'])

    if any(turbine['POW'][np.where(turbine['V'] == 0.0)] > 1e-2):
        logger.warning(
            "Oversmoothing detected with parameters eta=%f, Delta_v=%f, sigma=%f. "
            "Turbine generates energy at 0 m/s wind speeds.", eta, Delta_v, sigma)

    return turbine


def get_oedb_windturbineconfig(search=None, **search_params):
    """
    Download a windturbine configuration from the OEDB database.

    Download the configuration of a windturbine model from the OEDB database
    into the local 'windturbine_dir'.
    The OEDB database can be viewed here:
    https://openenergy-platform.org/dataedit/view/supply/wind_turbine_library
    (2019-07-22)
    Only one turbine configuration is downloaded at a time, if the
    search parameters yield an ambigious result, no data is downloaded.

    Parameters
    ----------
    search : int|str
        Smart search parameter, if int use as model id, if str look in name or turbine_type
    **search_params : dict
        Recognized arguments are 'id', 'name', 'turbine_type' and 'manufacturer'

    Returns
    -------
    turbineconfig : dict
        The turbine configuration in the format from 'atlite.ressource.get_turbineconf(name)'.

    Example
    -------
    >>> get_oedb_windturbineconfig(10)
    {'V': ..., 'POW': ..., ...}

    >>> get_oedb_windturbineconfig(name="E-53/800", manufacturer="Enercon")
    {'V': ..., 'POW': ..., ...}

    """

    # Parse information of different allowed 'turbine' values
    if isinstance(search, int):
        search_params.setdefault('id', search)
        search = None

    # Retrieve and cache OEDB turbine data
    OEDB_URL = 'https://openenergy-platform.org/api/v0/schema/supply/tables/wind_turbine_library/rows'

    # Cache turbine request locally
    global _oedb_turbines

    if _oedb_turbines is None:
        # Get the turbine list
        result = requests.get(OEDB_URL)

        # Convert JSON to dataframe for easier filtering
        # Only consider turbines with power curves available
        df = pd.DataFrame.from_dict(result.json())
        _oedb_turbines = df[df.has_power_curve]

    logger.info(
        "Searching turbine power curve in OEDB database using " +
        ", ".join(
            f"{k}='{v}'" for (
                k,
                v) in search_params.items()) +
        ".")

    # Working copy
    df = _oedb_turbines
    selector = True
    if search is not None:
        selector &= (df.name.str.contains(search, case=False) |
                     df.turbine_type.str.contains(search, case=False))
    if 'id' in search_params:
        selector &= df.id == int(search_params['id'])
    if 'name' in search_params:
        selector &= df.name.str.contains(search_params['name'], case=False)
    if 'turbine_type' in search_params:
        selector &= df.turbine_type.str.contains(
            search_params['name'], case=False)
    if 'manufacturer' in search_params:
        selector &= df.manufacturer.str.contains(
            search_params['manufacturer'], case=False)

    df = df.loc[selector]

    if len(df) < 1:
        raise RuntimeError("No turbine found.")
    elif len(df) > 1:
        raise RuntimeError(
            f"Provided information corresponds to {len(df)} turbines,"
            " use `id` for an unambiguous search.\n" +
            str(df[['id', 'manufacturer', 'turbine_type']])
        )

    # Convert to series for simpliticty
    ds = df.iloc[0]

    # convert power from kW to MW
    power = np.array(json.loads(ds.power_curve_values)) / 1e3

    hub_height = ds.hub_height

    if not hub_height:
        hub_height = 100
        logger.warning(
            "No hub_height defined in dataset. Manual clean-up required."
            "Assuming a hub_height of 100m for now.")
    elif isinstance(hub_height, str):
        hub_heights = [
            float(t) for t in re.split(
                r"\s*;\s*",
                hub_height.strip()) if t]

        if len(hub_heights) > 1:
            hub_height = np.mean(hub_heights, dtype=int)
            logger.warning(
                "Multiple values for hub_height in dataset (%s). "
                "Manual clean-up required. Using the averge %dm for now.",
                hub_heights,
                hub_height)
        else:
            hub_height = hub_heights[0]

    turbineconf = {
        "name": ds.turbine_type.strip(),
        "manufacturer": ds.manufacturer.strip(),
        "source": f"Original: {ds.source}. Via OEDB {OEDB_URL}",
        "hub_height": hub_height,
        "V": np.array(json.loads(ds.power_curve_wind_speeds)),
        "POW": power,
        "P": power.max()
    }

    # Cache in windturbines
    global windturbines
    charmap = str.maketrans('/- ', '___')
    name = "{manufacturer}_{name}".format(**turbineconf).translate(charmap)
    windturbines[name] = turbineconf

    return turbineconf


# Global caches
_oedb_turbines = None
windturbines = arrowdict(
    {p.stem: p for p in WINDTURBINE_DIRECTORY.glob("*.yaml")})
solarpanels = arrowdict(
    {p.stem: p for p in SOLARPANEL_DIRECTORY.glob("*.yaml")})



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import numpy as np
import xarray as xr
from numpy import sin, cos, deg2rad

def get_orientation(name, **params):
    '''
    Definitions:
        -`slope` is the angle between ground and panel.
        -`azimuth` is the clockwise angle from North.
            i.e. azimuth = 180 faces exactly South
    '''
    if isinstance(name, dict):
        params = name
        name = params.pop('name', 'constant')
    return getattr(sys.modules[__name__], 'make_{}'.format(name))(**params)


def make_latitude_optimal():
    """
    Returns an optimal tilt angle for the given ``lat``, assuming that
    the panel is facing towards the equator, using a simple method from [1].

    This method only works for latitudes between 0 and 50. For higher
    latitudes, a static 40 degree angle is returned.

    These results should be used with caution, but there is some
    evidence that tilt angle may not be that important [2].

    Function and documentation has been adapted from gsee [3].

    [1] http://www.solarpaneltilt.com/#fixed
    [2] http://dx.doi.org/10.1016/j.solener.2010.12.014
    [3] https://github.com/renewables-ninja/gsee/blob/master/gsee/pv.py

    Parameters
    ----------
    lat : float
        Latitude in degrees.
    """

    def latitude_optimal(lon, lat, solar_position):

        slope = np.empty_like(lat.values)

        below_25 = np.abs(lat.values) <= deg2rad(25)
        below_50 = np.abs(lat.values) <= deg2rad(50)

        slope[below_25] = 0.87 * np.abs(lat.values[below_25])
        slope[~below_25 & below_50] = (
            0.76 * np.abs(lat.values[~below_25 & below_50]) + deg2rad(0.31))
        slope[~below_50] = np.deg2rad(40.)

        # South orientation for panels on northern hemisphere and vice versa
        azimuth = np.where(lat.values < 0, 0, 180)

        return dict(slope=xr.DataArray(slope, coords=lat.coords),
                    azimuth=xr.DataArray(azimuth, coords=lat.coords))

    return latitude_optimal


def make_constant(slope, azimuth):
    slope = deg2rad(slope)
    azimuth = deg2rad(azimuth)

    def constant(lon, lat, solar_position):
        return dict(slope=slope, azimuth=azimuth)
    return constant


def make_latitude(azimuth=180):
    def latitude(lon, lat, solar_position):
        return dict(slope=lat, azimuth=azimuth)
    return latitude


def SurfaceOrientation(ds, solar_position, orientation):
    """
    Compute cos(incidence) for slope and panel azimuth

    References
    ----------
    [1] Sproul, A. B., Derivation of the solar geometric relationships using
    vector analysis, Renewable Energy, 32(7), 1187–1205 (2007).
    """

    lon = deg2rad(ds['lon'])
    lat = deg2rad(ds['lat'])

    orientation = orientation(lon, lat, solar_position)
    surface_slope = orientation['slope']
    surface_azimuth = orientation['azimuth']

    sun_altitude = solar_position['altitude']
    sun_azimuth = solar_position['azimuth']

    cosincidence = (sin(surface_slope) * cos(sun_altitude) *
                    cos(surface_azimuth - sun_azimuth) +
                    cos(surface_slope) * sin(sun_altitude))

    # fixup incidence angle: if the panel is badly oriented and the sun shines
    # on the back of the panel (incidence angle > 90degree), the irradiation
    # would be negative instead of 0; this is prevented here.
    cosincidence = cosincidence.clip(min=0)

    return xr.Dataset({'cosincidence': cosincidence,
                       'slope': surface_slope,
                       'azimuth': surface_azimuth})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for loading gebco data

"""

from rasterio.warp import Resampling
from pandas import to_numeric

import rasterio as rio
import xarray as xr

import logging
logger = logging.getLogger(__name__)

crs = 4326
features = {'height': ['height']}


def get_data_gebco_height(xs, ys, gebco_path):
    x, X = xs.data[[0, -1]]
    y, Y = ys.data[[0, -1]]

    dx = (X - x) / (len(xs) - 1)
    dy = (Y - y) / (len(ys) - 1)

    with rio.open(gebco_path) as dataset:
        window = dataset.window(x - dx/2, y - dy/2, X + dx/2, Y + dy/2)
        gebco = dataset.read(indexes=1,
                             window=window,
                             out_shape=(len(ys), len(xs)),
                             resampling=Resampling.average)
        gebco = gebco[::-1] # change inversed y-axis
        tags = dataset.tags(bidx=1)
        tags = {k: to_numeric(v, errors='ignore') for k, v in tags.items()}

    return xr.DataArray(gebco, coords=[("y", ys), ("x", xs)], name='height', attrs=tags)


def get_data(cutout, feature, tmpdir, **creation_parameters):
    """
    Get the gebco height data.

    Parameters
    ----------
    cutout : atlite.Cutout
    feature : str
        Takes no effect, only here for consistency with other dataset modules.
    tmpdir : str
        Takes no effect, only here for consistency with other dataset modules.
    **creation_parameters :
        Must include `gebco_path`.

    Returns
    -------
    xr.Dataset
    """
    if 'gebco_path' not in creation_parameters:
        logger.error('Argument "gebco_path" not defined')
    path = creation_parameters['gebco_path']

    coords = cutout.coords
    #assign time dimension even if not used
    return get_data_gebco_height(coords['x'], coords['y'], path)\
            .to_dataset()\
            .assign_coords(cutout.coords)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 11 11:15:41 2020

@author: fabian
"""

import os
import sys
import pytest
import urllib3
import geopandas as gpd
urllib3.disable_warnings()

import atlite
from atlite import Cutout
from xarray.testing import assert_allclose, assert_equal
import numpy as np


# %% Predefine tests for cutout

def all_notnull_test(cutout):
    """Test if no nan's in the prepared data occur"""
    assert cutout.data.notnull().all()


def prepared_features_test(cutout):
    """
    The prepared features series should contain all variables in cuttout.data
    """
    assert set(cutout.prepared_features) == set(cutout.data)


def update_feature_test(cutout, red):
    """atlite should be able to overwrite a feature."""
    red.data = cutout.data.drop_vars('influx_direct')
    red.prepare('influx', overwrite=True)
    assert_equal(red.data.influx_direct, cutout.data.influx_direct)


def merge_test(cutout, other, target_modules):
    merge = cutout.merge(other, compat='override')
    assert set(merge.module) == set(target_modules)


def wrong_recreation(cutout):
    with pytest.warns(UserWarning):
        Cutout(path=cutout.path, module='somethingelse')


def pv_test(cutout):
    '''
    Test the atlite.Cutout.pv function with different settings. Compare
    optimal orientation with flat orientation.
    '''

    orientation = {'slope': 0.0, 'azimuth': 0.0}
    cap_factor = cutout.pv(atlite.resource.solarpanels.CdTe, orientation)

    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    production = cutout.pv(atlite.resource.solarpanels.CdTe, orientation,
                        layout=cap_factor)

    assert production.notnull().all()
    assert (production.sel(time=TIME+ ' 00:00') == 0)

    cells = cutout.grid
    cells = cells.assign(regions=['lower']*200 + ['upper']*(len(cells)-200))
    shapes = cells.dissolve('regions')
    production, capacity = cutout.pv(atlite.resource.solarpanels.CdTe,
                                     orientation, layout=cap_factor,
                                     shapes=shapes, return_capacity=True)
    cap_per_region = (cells.assign(cap_factor=cap_factor.stack(spatial=['y','x']))
                          .groupby('regions').cap_factor.sum())

    assert all(cap_per_region.round(3) == capacity.round(3))

    # Now compare with optimal orienation
    cap_factor_opt = cutout.pv(atlite.resource.solarpanels.CdTe, 'latitude_optimal')

    assert cap_factor_opt.sum() > cap_factor.sum()

    production_opt = cutout.pv(atlite.resource.solarpanels.CdTe, 'latitude_optimal',
                            layout=cap_factor_opt)

    assert (production_opt.sel(time=TIME+ ' 00:00') == 0)

    assert production_opt.sum() > production.sum()



def wind_test(cutout):
    '''
    Test the atlite.Cutout.wind function with two different layouts.
    The standard layout proportional to the capacity factors must have a lower
    production than a layout proportionally to the capacity layout squared.
    '''
    cap_factor = cutout.wind(atlite.windturbines.Enercon_E101_3000kW)

    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    production = cutout.wind(atlite.windturbines.Enercon_E101_3000kW,
                          layout=cap_factor)

    assert production.notnull().all()
    assert production.sum() > 0

    # Now create a better layout with same amount of installed power
    better_layout = (cap_factor**2)/(cap_factor**2).sum() * cap_factor.sum()
    better_production = cutout.wind(atlite.windturbines.Enercon_E101_3000kW,
                                 layout=better_layout)

    assert better_production.sum() > production.sum()


def runoff_test(cutout):
    '''
    Test the atlite.Cutout.runoff function.

    First check if the total of all capacity factors is not null.
    Then compare the runoff at sites which belong to the lower (altitude) half
    of the map, with the runoff at higher sites. The runoff at the lower sites
    (mostly at sea level) should have a smaller capacity factor total and
    production.
    '''
    cap_factor = cutout.runoff()
    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    height = cutout.data.height.load()
    q = np.quantile(height, 0.5)
    lower_area = height <= q
    higher_area = height > q
    assert cap_factor.where(lower_area).sum() < cap_factor.where(higher_area).sum()

    low_level_prod = cutout.runoff(layout=cap_factor.where(lower_area, 0))
    high_level_prod = cutout.runoff(layout=cap_factor.where(higher_area, 0))
    assert low_level_prod.sum() < high_level_prod.sum()


# I don't understand the problems with the crs and projection here leaving this
# out:

# def test_hydro():
#     plants = pd.DataFrame({'lon' : [x0, x1],
#                            'lat': [y0, y1]})
#     basins = gpd.GeoDataFrame(dict(geometry=[cutout_era5.grid_cells[0], cutout_era5.grid_cells[-1]],
#                                    HYBAS_ID = [0,1],
#                                    DIST_MAIN = 10,
#                                    NEXT_DOWN = None), index=[0,1], crs=dict(proj="aea"))
#     cutout_era5.hydro(plants, basins)


# %% Prepare cutouts to test


TIME = '2013-01-01'
BOUNDS = (-4, 56, 1.5, 61)
SARAH_DIR = os.getenv('SARAH_DIR', '/home/vres/climate-data/sarah_v2')


@pytest.fixture(scope='session')
def cutout_era5(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("era5")
    cutout = Cutout(path=tmp_path / "era5", module="era5", bounds=BOUNDS, time=TIME)
    cutout.prepare()
    return cutout


@pytest.fixture(scope='session')
def cutout_era5_coarse(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("era5_coarse")
    cutout = Cutout(path=tmp_path / "era5", module="era5", bounds=BOUNDS,
                    time=TIME, dx=0.5, dy=0.7)
    cutout.prepare()
    return cutout


@pytest.fixture(scope='session')
def cutout_era5_weird(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("era5_weird")
    cutout = Cutout(path=tmp_path / "era5", module="era5", bounds=BOUNDS,
                    time=TIME, dx=0.132, dy=0.32)
    cutout.prepare()
    return cutout


@pytest.fixture(scope='session')
def cutout_era5_reduced(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("era5_red")
    cutout = Cutout(path=tmp_path / "era5", module="era5", bounds=BOUNDS, time=TIME)
    return cutout


@pytest.fixture(scope='session')
def cutout_sarah(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("sarah")
    cutout = Cutout(path=tmp_path / "sarah", module=["sarah", "era5"],
                    bounds=BOUNDS, time=TIME, sarah_dir=SARAH_DIR)
    cutout.prepare()
    return cutout


@pytest.fixture(scope='session')
def cutout_sarah_fine(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("sarah_coarse")
    cutout = Cutout(path=tmp_path / "sarah", module="sarah", bounds=BOUNDS,
                    time=TIME, dx=0.05, dy=0.05, sarah_dir=SARAH_DIR)
    cutout.prepare()
    return cutout


@pytest.fixture(scope='session')
def cutout_sarah_weird(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("sarah_weird")
    cutout = Cutout(path=tmp_path / "sarah", module="sarah", bounds=BOUNDS,
                    time=TIME, dx=0.132, dy=0.32, sarah_dir=SARAH_DIR)
    cutout.prepare()
    return cutout






class TestERA5:
    @staticmethod
    def test_data_module_arguments_era5(cutout_era5):
        """
        All data variables should have an attribute to which module thay belong
        """
        for v in cutout_era5.data:
            assert cutout_era5.data.attrs['module'] == 'era5'

    @staticmethod
    def test_all_non_na_era5(cutout_era5):
        """Every cells should have data."""
        assert np.isfinite(cutout_era5.data).all()

    @staticmethod
    def test_all_non_na_era5_coarse(cutout_era5_coarse):
        """Every cells should have data."""
        assert np.isfinite(cutout_era5_coarse.data).all()

    @staticmethod
    @pytest.mark.skipif(os.name == 'nt',
                        reason="This test breaks on windows machine on travis"
                        " due to unknown reasons.")
    def test_all_non_na_era5_weird(cutout_era5_weird):
        """Every cells should have data."""
        assert np.isfinite(cutout_era5_weird.data).all()


    @staticmethod
    def test_dx_dy_preservation_era5(cutout_era5):
        """The coordinates should be the same after preparation."""
        assert np.allclose(np.diff(cutout_era5.data.x), 0.25)
        assert np.allclose(np.diff(cutout_era5.data.y), 0.25)

    @staticmethod
    def test_dx_dy_preservation_era5_coarse(cutout_era5_coarse):
        """The coordinates should be the same after preparation."""
        assert np.allclose(np.diff(cutout_era5_coarse.data.x),
                           cutout_era5_coarse.data.attrs['dx'])
        assert np.allclose(np.diff(cutout_era5_coarse.data.y),
                           cutout_era5_coarse.data.attrs['dy'])

    @staticmethod
    @pytest.mark.skipif(os.name == 'nt',
                        reason="This test breaks on windows machine on travis"
                        " due to unknown reasons.")
    def test_dx_dy_preservation_era5_weird(cutout_era5_weird):
        """The coordinates should be the same after preparation."""
        assert np.allclose(np.diff(cutout_era5_weird.data.x),
                           cutout_era5_weird.data.attrs['dx'])
        assert np.allclose(np.diff(cutout_era5_weird.data.y),
                           cutout_era5_weird.data.attrs['dy'])


    @staticmethod
    def test_compare_with_get_data_era5(cutout_era5, tmp_path):
        """
        The prepared data should be exactly the same as from the low level function
        """
        influx = atlite.datasets.era5.get_data(cutout_era5, 'influx', tmpdir=tmp_path)
        assert_allclose(influx.influx_toa, cutout_era5.data.influx_toa, atol=1e-5, rtol=1e-5)

    @staticmethod
    def test_prepared_features_era5(cutout_era5):
        return prepared_features_test(cutout_era5)

    @staticmethod
    @pytest.mark.skipif(sys.platform == "win32",
                        reason='NetCDF update not working on windows')
    def test_update_feature_era5(cutout_era5, cutout_era5_reduced):
        return update_feature_test(cutout_era5, cutout_era5_reduced)

    @staticmethod
    def test_wrong_loading(cutout_era5):
        wrong_recreation(cutout_era5)

    @staticmethod
    def test_pv_era5(cutout_era5):
        return pv_test(cutout_era5)

    @staticmethod
    def test_wind_era5(cutout_era5):
        return wind_test(cutout_era5)

    @staticmethod
    def test_runoff_era5(cutout_era5):
        return runoff_test(cutout_era5)


@pytest.mark.skipif(not os.path.exists(SARAH_DIR),
                    reason="'sarah_dir' is not a valid path")
class TestSarah:

    @staticmethod
    def test_all_non_na_sarah(cutout_sarah):
        """Every cells should have data."""
        assert np.isfinite(cutout_sarah.data).all()

    @staticmethod
    def test_all_non_na_sarah_fine(cutout_sarah_fine):
        """Every cells should have data."""
        assert np.isfinite(cutout_sarah_fine.data).all()

    @staticmethod
    def test_all_non_na_sarah_weird(cutout_sarah_weird):
        """Every cells should have data."""
        assert np.isfinite(cutout_sarah_weird.data).all()

    @staticmethod
    def test_dx_dy_preservation_sarah(cutout_sarah):
        """The coordinates should be the same after preparation."""
        assert np.allclose(np.diff(cutout_sarah.data.x), 0.25)
        assert np.allclose(np.diff(cutout_sarah.data.y), 0.25)


    @staticmethod
    def test_prepared_features_sarah(cutout_sarah):
        return prepared_features_test(cutout_sarah)

    @staticmethod
    def test_merge(cutout_sarah, cutout_era5):
        return merge_test(cutout_sarah, cutout_era5, ['sarah', 'era5'])

    @staticmethod
    def test_pv_sarah(cutout_sarah):
        return pv_test(cutout_sarah)


    @staticmethod
    def test_wind_sarah(cutout_sarah):
        return wind_test(cutout_sarah)

    @staticmethod
    def test_runoff_sarah(cutout_sarah):
        return runoff_test(cutout_sarah)



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later


import numpy as np

# Huld model was copied from gsee -- global solar energy estimator
# by Stefan Pfenninger
# https://github.com/renewables-ninja/gsee/blob/master/gsee/pv.py


def _power_huld(irradiance, t_amb, pc):
    """
    AC power per capacity predicted by Huld model, based on W/m2 irradiance.

    Maximum power point tracking is assumed.

    [1] Huld, T. et al., 2010. Mapping the performance of PV modules,
        effects of module type and data averaging. Solar Energy, 84(2),
        p.324-338. DOI: 10.1016/j.solener.2009.12.002
    """

    # normalized module temperature
    T_ = (pc['c_temp_amb'] * t_amb + pc['c_temp_irrad']
          * irradiance) - pc['r_tmod']

    # normalized irradiance
    G_ = irradiance / pc['r_irradiance']

    log_G_ = np.log(G_.where(G_>0))
    # NB: np.log without base implies base e or ln
    eff = (1 + pc['k_1'] * log_G_ + pc['k_2'] * (log_G_) ** 2 +
           T_ * (pc['k_3'] + pc['k_4'] * log_G_ + pc['k_5'] * log_G_ ** 2) +
           pc['k_6'] * (T_ ** 2))

    eff = eff.fillna(0.).clip(min=0)

    return G_ * eff * pc.get('inverter_efficiency', 1.)


def _power_bofinger(irradiance, t_amb, pc):
    """
    AC power per capacity predicted by bofinger model, based on W/m2 irradiance.

    Maximum power point tracking is assumed.

    [2] Hans Beyer, Gerd Heilscher and Stefan Bofinger, 2004. A robust model
    for the MPP performance of different types of PV-modules applied for the
    performance check of grid connected systems.
    """

    fraction = (pc['NOCT'] - pc['Tamb']) / pc['Intc']

    eta_ref = (
        pc['A'] +
        pc['B'] *
        irradiance +
        pc['C'] *
        np.log(irradiance.where(irradiance!=0)))
    eta = (eta_ref *
           (1. + pc['D'] * (fraction * irradiance + (t_amb - pc['Tstd']))) /
           (1. + pc['D'] * fraction / pc['ta'] * eta_ref * irradiance)
           ).fillna(0)

    capacity = (pc['A'] + pc['B'] * 1000. + pc['C'] * np.log(1000.)) * 1e3
    power = irradiance * eta * (pc.get('inverter_efficiency', 1.) / capacity)
    power = power.where(irradiance >= pc['threshold'], 0)
    return power.rename('AC power')


def SolarPanelModel(ds, irradiance, pc):
    model = pc.get('model', 'huld')

    if model == 'huld':
        return _power_huld(irradiance, ds['temperature'], pc)
    elif model == 'bofinger':
        return _power_bofinger(irradiance, ds['temperature'], pc)
    else:
        AssertionError("Unknown panel model: {}".format(model))



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Base class for Atlite.
"""

# There is a binary incompatibility between the pip wheels of netCDF4 and
# rasterio, which leads to the first one to work correctly while the second
# loaded one fails by loading netCDF4 first, we ensure that most of atlite's
# functionality works fine, even when the pip wheels have been used, only for
# resampling the sarah dataset it is important to use conda.
# Refer to
# https://github.com/pydata/xarray/issues/2535,
# https://github.com/rasterio/rasterio-wheels/issues/12

import xarray as xr
import pandas as pd
import numpy as np
import dask
import rasterio as rio
import geopandas as gpd
from tempfile import mktemp
from numpy import atleast_1d, append
from warnings import warn
from shapely.geometry import box
from pathlib import Path
from pyproj import CRS


from .utils import CachedAttribute
from .data import cutout_prepare, available_features
from .gis import get_coords, compute_indicatormatrix, compute_availabilitymatrix
from .convert import (convert_and_aggregate, heat_demand, hydro, temperature,
                      wind, pv, runoff, solar_thermal, soil_temperature)
from .datasets import modules as datamodules

import logging
logger = logging.getLogger(__name__)


class Cutout:
    """
    Cutout base class.

    This class builds the starting point for most atlite functionalities.
    """

    def __init__(self, path, **cutoutparams):
        """
        Provide an Atlite cutout object.

        Create a cutout object to use atlite operations on it. Based on the
        provided parameters, atlite first checks whether this cutout already
        exists on disk and if yes, loads this cutout.

        If the cutout does not yet exist on disk, then atlite creates an
        "unprepared" cutout object. This does not yet contain the full data.
        The process of preparing (loading the data) can then be started with
        `cutout.prepare()`.

        Parameters
        ----------
        path : str | path-like
            NetCDF from which to load or where to store the cutout.
        module : str or list
            The dataset(s) which works as a basis for the cutout. Available
            modules are "era5", "sarah" and "gebco".
            This is necessary when building a new cutout.
            If more than one module is given, their order determines how atlite
            fills up missing features when preparing the cutout with
            `Cutout.prepare()`. For example `influx_diffuse` is provided by
            the `sarah` and the `era5` module. Prioritizing sarah and setting
            module=['sarah', 'era5'] will load `influx_diffuse` from the sarah
            module and ignoring the era5 'influx_diffuse' data.
        time : str | slice
            Time range to include in the cutout, e.g. "2011" or
            ("2011-01-05", "2011-01-25")
            This is necessary when building a new cutout.
        bounds : GeoSeries.bounds | DataFrame, optional
            The outer bounds of the cutout or as a DataFrame
            containing (min.long, min.lat, max.long, max.lat).
        x : slice, optional
            Outer longitudinal bounds for the cutout (west, east).
        y : slice, optional
            Outer latitudinal bounds for the cutout (south, north).
        dx : float, optional
            Step size of the x coordinate. The default is 0.25.
        dy : float, optional
            Step size of the y coordinate. The default is 0.25.
        dt : str, optional
            Frequency of the time coordinate. The default is 'h'. Valid are all
            pandas offset aliases.
        chunks : dict
            Chunks when opening netcdf files. For cutout preparation recommand
            to chunk only along the time dimension. Defaults to {'time': 20}
        data : xr.Dataset
            User provided cutout data. Save the cutout using `Cutout.to_file()`
            afterwards.

        Other Parameters
        ----------------
        sanitize : bool, default True
            Whether to sanitize the data when preparing the cutout. Takes
            effect for 'era5' data loading.
        sarah_dir : str, Path
            Directory of on-disk sarah data. This must be given when using the
            sarah module.
        sarah_interpolate : bool, default True
            Whether to interpolate NaN's in the SARAH data. This takes effect for
            sarah data which has missing data for areas where dawn and
            nightfall happens (ca. 30 min gap).
        gebco_path: str
            Path to find the gebco netcdf file. Only necessary when including
            the gebco module.
        parallel : bool, default False
            Whether to open dataset in parallel mode. Take effect for all
            xr.open_mfdataset usages.

        """
        name = cutoutparams.get("name", None)
        cutout_dir = cutoutparams.get("cutout_dir", None)
        if cutout_dir or name or Path(path).is_dir():
            raise ValueError(
                "Old style format not supported. You can migrate the old "
                "cutout directory using the function "
                "`atlite.utils.migrate_from_cutout_directory()`. The argument "
                "`cutout_dir` and `name` have been deprecated in favour of `path`.")

        path = Path(path).with_suffix(".nc")
        chunks = cutoutparams.pop('chunks', {'time': 100})
        storable_chunks = {f'chunksize_{k}': v for k, v in (chunks or {}).items()}

        # Backward compatibility for xs, ys, months and years
        if {'xs', 'ys'}.intersection(cutoutparams):
            warn(
                "The arguments `xs` and `ys` have been deprecated in favour of "
                "`x` and `y`", DeprecationWarning)
            if 'xs' in cutoutparams:
                cutoutparams['x'] = cutoutparams.pop('xs')
            if 'ys' in cutoutparams:
                cutoutparams['y'] = cutoutparams.pop('ys')

        if {'years', 'months'}.intersection(cutoutparams):
            warn("The arguments `years` and `months` have been deprecated in "
                 "favour of `time`", DeprecationWarning)
            assert 'years' in cutoutparams
            months = cutoutparams.pop("months", slice(1, 12))
            years = cutoutparams.pop("years")
            cutoutparams["time"] = slice(f"{years.start}-{months.start}",
                                         f"{years.stop}-{months.stop}")

        # Three cases. First, cutout exists -> take the data.
        # Second, data is given -> take it. Third, else -> build a new cutout
        if path.is_file():
            data = xr.open_dataset(str(path), chunks=chunks)
            data.attrs.update(storable_chunks)
            if cutoutparams:
                warn(f'Arguments {", ".join(cutoutparams)} are ignored, since '
                     'cutout is already built.')
        elif 'data' in cutoutparams:
            data = cutoutparams.pop('data')
        else:
            logger.info(f"Building new cutout {path}")

            if 'bounds' in cutoutparams:
                x1, y1, x2, y2 = cutoutparams.pop('bounds')
                cutoutparams.update(x=slice(x1, x2), y=slice(y1, y2))

            try:
                x = cutoutparams.pop('x')
                y = cutoutparams.pop('y')
                time = cutoutparams.pop('time')
                module = cutoutparams.pop('module')
            except KeyError as exc:
                raise TypeError("Arguments 'time' and 'module' must be "
                                "specified. Spatial bounds must either be "
                                "passed via argument 'bounds' or 'x' and 'y'.") from exc

            # TODO: check for dx, dy, x, y fine with module requirements
            coords = get_coords(x, y, time, **cutoutparams)

            attrs = {'module': module, 'prepared_features': [],
                     **storable_chunks, **cutoutparams}
            data = xr.Dataset(coords=coords, attrs=attrs)

        # Check compatibility of CRS
        modules = atleast_1d(data.attrs.get('module'))
        crs = set(CRS(datamodules[m].crs) for m in modules)
        assert len(crs) == 1, f'CRS of {module} not compatible'

        self.path = path
        self.data = data


    @property
    def name(self):
        return self.path.stem

    @property
    def module(self):
        return self.data.attrs.get('module')

    @property
    def crs(self):
        return CRS(datamodules[atleast_1d(self.module)[0]].crs)

    @property
    def available_features(self):
        return available_features(self.module)

    @property
    def chunks(self):
        chunks = {k.lstrip('chunksize_'): v for k, v in self.data.attrs.items()
                  if k.startswith('chunksize_')}
        return None if chunks == {} else chunks

    @property
    def coords(self):
        return self.data.coords


    @property
    def meta(self):
        warn("The `meta` attribute is deprecated in favour of direct "
             "access to `data`", DeprecationWarning)
        return xr.Dataset(self.coords, attrs=self.data.attrs)

    @property
    def shape(self):
        return len(self.coords["y"]), len(self.coords["x"])

    @property
    def extent(self):
        """Total extent of the area covered by the cutout (x, X, y, Y)."""
        xs, ys = self.coords['x'].values, self.coords['y'].values
        dx , dy = self.dx, self.dy
        return np.array([xs[0]-dx/2, xs[-1]+dx/2, ys[0]-dy/2, ys[-1]+dy/2])

    @property
    def bounds(self):
        """Total bounds of the area covered by the cutout (x, y, X, Y)."""
        return self.extent[[0,2,1,3]]

    @property
    def transform(self):
        """Get the affine transform of the cutout. """
        return rio.Affine(self.dx, 0, self.coords['x'].values[0] - self.dx/2,
                          0, self.dy, self.coords['y'].values[0] - self.dy/2)

    @property
    def transform_r(self):
        """Get the affine transform of the cutout with reverse y-order."""
        return rio.Affine(self.dx, 0, self.coords['x'].values[0] - self.dx/2,
                          0, -self.dy, self.coords['y'].values[-1] + self.dy/2)

    @property
    def dx(self):
        x = self.coords['x']
        return round((x[-1] - x[0]).item() / (x.size - 1), 8)

    @property
    def dy(self):
        y = self.coords['y']
        return round((y[-1] - y[0]).item() / (y.size - 1), 8)

    @property
    def dt(self):
        return pd.infer_freq(self.coords['time'].to_index())

    @property
    def prepared(self):
        return (self.prepared_features.sort_index()
                .equals(self.available_features.sort_index()))

    @property
    def prepared_features(self):
        index = [(self.data[v].attrs['module'], self.data[v].attrs['feature'])
                 for v in self.data]
        index = pd.MultiIndex.from_tuples(index, names=['module', 'feature'])
        return pd.Series(list(self.data), index)

    def grid_coordinates(self):
        warn("The function `grid_coordinates` has been deprecated in favour of "
             "`grid`", DeprecationWarning)
        logger.warning("The order of elements returned by `grid_coordinates` changed. "
                       "Check the output of your workflow for correctness.")
        return self.grid[['x', 'y']].values

    def grid_cells(self):
        warn("The function `grid_cells` has been deprecated in favour of `grid`",
             DeprecationWarning)
        logger.warning("The order of elements in `grid_cells` changed. "
                       "Check the output of your workflow for correctness.")
        return self.grid.geometry.to_list()


    @CachedAttribute
    def grid(self):
        xs, ys = np.meshgrid(self.coords["x"], self.coords["y"])
        coords = np.asarray((np.ravel(xs), np.ravel(ys))).T
        span = (coords[self.shape[1] + 1] - coords[0]) / 2
        cells = [box(*c) for c in np.hstack((coords - span, coords + span))]
        return gpd.GeoDataFrame({'x': coords[:, 0], 'y': coords[:, 1],
                                 'geometry': cells,}, crs=self.crs)


    def sel(self, path=None, bounds=None, buffer=0, **kwargs):
        '''
        Select parts of the cutout.

        Parameters
        ----------
        path : str | path-like
            File where to store the sub-cutout. Defaults to a temporary file.
        bounds : GeoSeries.bounds | DataFrame, optional
            The outer bounds of the cutout or as a DataFrame
            containing (min.long, min.lat, max.long, max.lat).
        buffer : float, optional
            Buffer around the bounds. The default is 0.
        **kwargs :
            Passed to `xr.Dataset.sel` for data selection.

        Returns
        -------
        selected : Cutout
            Selected cutout.

        '''
        if path is None:
            path = mktemp(prefix=f"{self.path.stem}-", suffix=self.path.suffix,
                          dir=self.path.parent)

        if bounds is not None:
            if buffer > 0:
                bounds = box(*bounds).buffer(buffer).bounds
            x1, y1, x2, y2 = bounds
            kwargs.update(x=slice(x1, x2), y=slice(y1, y2))
        data = self.data.sel(**kwargs)
        return Cutout(path, data=data)


    def merge(self, other, path=None, **kwargs):
        '''
        Merge two cutouts into a single cutout.

        Parameters
        ----------
        other : atlite.Cutout
            Other cutout to merge.
        path : str | path-like
            File where to store the merged cutout. Defaults to a temporary file.
        **kwargs
            Keyword arguments passed to `xarray.merge()`.

        Returns
        -------
        merged : Cutout
            Merged cutout.

        '''
        assert isinstance(other, Cutout)

        if path is None:
            path = mktemp(prefix=f"{self.path.stem}-", suffix=self.path.suffix,
                          dir=self.path.parent)

        attrs = {**self.data.attrs, **other.data.attrs}
        attrs['module'] = list(set(append(*atleast_1d(self.module, other.module))))
        features = self.prepared_features.index.unique('feature')
        otherfeatures = other.prepared_features.index.unique('feature')
        attrs['prepared_features'] = list(features.union(otherfeatures))

        data = self.data.merge(other.data, **kwargs).assign_attrs(**attrs)

        return Cutout(path, data=data)


    def to_file(self, fn=None):
        '''
        Save cutout to a netcdf file.

        Parameters
        ----------
        fn : str | path-like
            File name where to store the cutout, defaults to `cutout.path`.

        '''
        if fn is None:
            fn = self.path
        self.data.to_netcdf(fn)


    def __repr__(self):
        start = np.datetime_as_string(self.coords['time'].values[0], unit='D')
        end = np.datetime_as_string(self.coords['time'].values[-1], unit='D')
        return ('<Cutout "{}">\n'
                ' x = {:.2f} ⟷ {:.2f}, dx = {:.2f}\n'
                ' y = {:.2f} ⟷ {:.2f}, dy = {:.2f}\n'
                ' time = {} ⟷ {}, dt = {}\n'
                ' module = {}\n'
                ' prepared_features = {}'
                .format(self.name, self.coords['x'].values[0],
                        self.coords['x'].values[-1], self.dx,
                        self.coords['y'].values[0],
                        self.coords['y'].values[-1], self.dy,
                        start, end, self.dt,
                        self.module,
                        list(self.prepared_features.index.unique('feature'))))


    def indicatormatrix(self, shapes, shapes_crs=4326):
        """
        Compute the indicatormatrix.

        The indicatormatrix I[i,j] is a sparse representation of the ratio
        of the area in orig[j] lying in dest[i], where orig and dest are
        collections of polygons, i.e.

        A value of I[i,j] = 1 indicates that the shape orig[j] is fully
        contained in shape dest[j].

        Note that the polygons must be in the same crs.

        Parameters
        ---------
        shapes : Collection of shapely polygons

        Returns
        -------
        I : sp.sparse.lil_matrix
          Indicatormatrix
        """
        return compute_indicatormatrix(self.grid, shapes, self.crs, shapes_crs)


    def uniform_layout(self):
        """Get a uniform capacity layout for all grid cells."""
        return xr.DataArray(1, [self.coords['y'], self.coords['x']])


    def layout_from_capacity_list(self, data, col='Capacity'):
        """
        Get a capacity layout aligned to the cutout based on a capacity list.

        Parameters
        ----------
        data : pandas.DataFrame
            Capacity list with columns 'x', 'y' and col. Each capacity entry
            is added to the grid cell intersecting with the coordinate (x,y).
        col : str, optional
            Name of the column with capacity values. The default is 'Capacity'.

        Returns
        -------
        xr.DataArray
            Capacity layout with dimensions 'x' and 'y' indicating the total
            capacity placed within one grid cell.

        Example
        -------
        >>> import atlite
        >>> import powerplantmatching as pm

        >>> data = pm.data.OPSD_VRE_country('DE')
        >>> data = (data.query('Fueltype == "Solar"')
                    .rename(columns={'lon':'x', 'lat':'y'}))

        >>> cutout = atlite.Cutout('Germany', x = slice(-5, 15), y = slice(40, 55),
                       time='2013-06-01', module='era5')
        >>> cutout.prepare(features=['influx', 'temperature'])
        >>> layout = cutout.layout_from_capacity_list(data)
        >>> pv = cutout.pv('CdTe', 'latitude_optimal', layout=layout)
        >>> pv.plot()

        """
        with dask.config.set(**{'array.slicing.split_large_chunks': False}):
            nearest = (self.uniform_layout().chunk()
                       .sel({'x': data.x.values, 'y': data.y.values}, 'nearest'))

        data = (data.assign(x=nearest.x.data, y=nearest.y.data)
                    .groupby(['y', 'x'])[col].sum())
        return data.to_xarray().reindex_like(self.data).fillna(0)


    availabilitymatrix = compute_availabilitymatrix

    # Preparation functions

    prepare = cutout_prepare

    # Conversion and aggregation functions

    convert_and_aggregate = convert_and_aggregate

    heat_demand = heat_demand

    temperature = temperature

    soil_temperature = soil_temperature

    solar_thermal = solar_thermal

    wind = wind

    pv = pv

    runoff = runoff

    hydro = hydro



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  6 15:23:13 2020

@author: fabian
"""

# IDEAS for tests

import pytest
import pandas as pd
import atlite
from atlite import Cutout
from xarray.testing import assert_allclose, assert_equal
import numpy as np


TIME = '2013-01-01'
X0 = -4.
Y0 = 56.
X1 = 1.5
Y1 = 61.


@pytest.fixture
def ref():
    return Cutout(path="creation_ref", module="era5", bounds=(X0, Y0, X1, Y1), time=TIME)


def test_odd_bounds_coords(ref):
    cutout = Cutout(path="odd_bounds", module="era5", time=TIME,
                    bounds=(X0-0.1, Y0-0.02, X1+0.03, Y1+0.13))
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_xy_coords(ref):
    cutout = Cutout(path="xy", module="era5", time=TIME,
                    x=slice(X0, X1), y = slice(Y0, Y1))
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_xy_reversed_coords(ref):
    cutout = Cutout(path="xy_r", module="era5", time=TIME,
                    x=slice(X1, X0), y = slice(Y1, Y0))
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())



def test_time_sclice_coords(ref):
    cutout = Cutout(path="time_slice", module="era5",
                    time=slice('2013-01-01', '2013-01-01'),
                    x=slice(X0, X1), y = slice(Y0, Y1))
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())



def test_dx_dy_dt():
    """
    Test the properties dx, dy, dt of atlite.Cutout. The spatial resolution
    can be changed through the creation_params dx and dy, the time resolution
    is hard coded (deep in the modules...) to one hour.
    """
    dx = 0.5
    dy = 1
    cutout = Cutout(path="resolution", module="era5",
                    time=slice('2013-01-01', '2013-01-01'),
                    x=slice(X0, X1), y = slice(Y0, Y1),
                    dx=dx, dy=dy)
    assert dx == cutout.dx
    assert dy == cutout.dy
    assert 'H' == cutout.dt


def test_available_features(ref):
    modules = ref.available_features.index.unique('module')
    assert len(modules) == 1
    assert modules[0] == 'era5'

    cutout = Cutout(path="sarah_first", module=['sarah', 'era5'],
                    time=slice('2013-01-01', '2013-01-01'),
                    x=slice(X0, X1), y = slice(Y0, Y1))
    modules = cutout.available_features.index.unique('module')
    assert len(modules) == 2
    assert len(cutout.available_features) > len(ref.available_features)


def test_sel(ref):
    cutout = ref.sel(x=slice(X0+2, X1-1), y=slice(Y0+1, Y1-2))
    assert cutout.coords['x'][0] - ref.coords['x'][0] == 2
    assert cutout.coords['y'][-1] - ref.coords['y'][-1] == -2




# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module containing specific operations for creating cutouts from the SARAH2 dataset.
"""

from ..gis import regrid
from rasterio.warp import Resampling
import os
import glob
import pandas as pd
import numpy as np
import xarray as xr
from functools import partial

import logging
logger = logging.getLogger(__name__)


# Model, CRS and Resolution Settings
crs = 4326
dx = 0.05
dy = 0.05
dt = '30min'
features = {'influx': ['influx_direct', 'influx_diffuse',]}
static_features = {}


def get_filenames(sarah_dir, coords):
    """
    Get all files in directory `sarah_dir` relevent for coordinates `coords`.

    This function parses all files in the sarah directory which lay in the time
    span of the coordinates.

    Parameters
    ----------
    sarah_dir : str
    coords : atlite.Cutout.coords

    Returns
    -------
    pd.DataFrame with two columns `sis` and `sid` for and timeindex for all
    relevant files.
    """
    def _filenames_starting_with(name):
        pattern = os.path.join(sarah_dir, "**", f"{name}*.nc")
        files = pd.Series(glob.glob(pattern, recursive=True))
        assert not files.empty, (f"No files found at {pattern}. Make sure "
                                 "sarah_dir points to the correct directory!")

        files.index = pd.to_datetime(files.str.extract(r"SI.in(\d{8})",
                                                       expand=False))
        return files.sort_index()
    files = pd.concat(dict(sis=_filenames_starting_with("SIS"),
                           sid=_filenames_starting_with("SID")),
                      join="inner", axis=1)

    start = coords['time'].to_index()[0]
    end = coords['time'].to_index()[-1]

    if (start < files.index[0]) or (end.date() > files.index[-1]):
        logger.error(f"Files in {sarah_dir} do not cover the whole time span:"
                     f"\n\t{start} until {end}")

    return files.loc[(files.index >= start) & (files.index <= end)].sort_index()



def interpolate(ds, dim='time'):
    """
    Interpolate NaNs in a dataset along a chunked dimension.

    This function is similar to xr.Dataset.interpolate_na but can
    be used for interpolating along a chunked dimensions (default 'time'').
    As the sarah data has mulitple NaN's in the areas of dawn and nightfall
    and the data is per default chunked along the time axis, use this function
    to interpolate.
    """
    def _interpolate1d(y):
        nan = np.isnan(y)
        if nan.all() or not nan.any():
            return y

        def x(z): return z.nonzero()[0]
        y = np.array(y)
        y[nan] = np.interp(x(nan), x(~nan), y[~nan])
        return y

    def _interpolate(a):
        return a.map_blocks(partial(np.apply_along_axis, _interpolate1d, -1),
                            dtype=a.dtype)

    data_vars = ds.data_vars.values() if isinstance(ds, xr.Dataset) else (ds,)
    dtypes = {da.dtype for da in data_vars}
    assert len(dtypes) == 1, \
        "interpolate only supports datasets with homogeneous dtype"

    return xr.apply_ufunc(_interpolate, ds,
                          input_core_dims=[[dim]],
                          output_core_dims=[[dim]],
                          output_dtypes=[dtypes.pop()],
                          output_sizes={dim: len(ds.indexes[dim])},
                          dask='allowed',
                          keep_attrs=True)


def as_slice(bounds, pad=True):
    """Convert coordinate bounds to slice and pad by 0.01."""
    if not isinstance(bounds, slice):
        bounds = bounds + (-.01, .01)
        bounds = slice(*bounds)
    return bounds


def hourly_mean(ds):
    """
    Resample time data to one hour frequency.

    In contrast to the standard xarray resample function this preserves chunks
    sizes along the time dimension.
    """
    ds1 = ds.isel(time=slice(None, None, 2))
    ds2 = ds.isel(time=slice(1, None, 2))
    ds2 = ds2.assign_coords(time=ds2.indexes['time'] - pd.Timedelta(30, 'm'))
    ds = ((ds1 + ds2)/2)
    ds.attrs = ds1.attrs
    for v in ds.variables:
        ds[v].attrs = ds1[v].attrs
    return ds


def get_data(cutout, feature, tmpdir, lock=None, **creation_parameters):
    """
    Load stored SARAH data and reformat to matching the given cutout.

    This function loads and resamples the stored SARAH data for a given
    `atlite.Cutout`.

    Parameters
    ----------
    cutout : atlite.Cutout
    feature : str
        Name of the feature data to retrieve. Must be in
        `atlite.datasets.sarah.features`
    **creation_parameters :
        Mandatory arguments are:
            * 'sarah_dir', str. Directory of the stored SARAH data.
        Possible arguments are:
            * 'parallel', bool. Whether to load stored files in parallel
            mode. Default is False.
            * 'sarah_interpolate', bool. Whether to interpolate areas of dawn
            and nightfall. This might slow down the loading process if only
            a few cores are available. Default is True.

    Returns
    -------
    xarray.Dataset
        Dataset of dask arrays of the retrieved variables.

    """
    assert cutout.dt in ("30min", "30T", 'h', 'H', '1h', '1H')

    coords = cutout.coords
    chunks = cutout.chunks

    sarah_dir = creation_parameters['sarah_dir']
    creation_parameters.setdefault('parallel', False)
    creation_parameters.setdefault('sarah_interpolate', True)

    files = get_filenames(sarah_dir, coords)
    open_kwargs = dict(chunks=chunks, parallel=creation_parameters['parallel'])
    ds_sis = xr.open_mfdataset(files.sis, combine='by_coords', **open_kwargs)
    ds_sid = xr.open_mfdataset(files.sid, combine='by_coords', **open_kwargs)
    ds = xr.merge([ds_sis, ds_sid])
    ds = ds.sel(lon=as_slice(cutout.extent[:2]), lat=as_slice(cutout.extent[2:]))
    # fix float (im)precission
    ds = ds.assign_coords(lon=ds.lon.astype(float).round(4),
                          lat=ds.lat.astype(float).round(4))

    # Interpolate, resample and possible regrid
    if creation_parameters['sarah_interpolate']:
        ds = interpolate(ds)
    else:
        ds = ds.fillna(0)

    if cutout.dt not in ['30min', '30T']:
        ds = hourly_mean(ds)

    if (cutout.dx != dx) or (cutout.dy != dy):
        ds = regrid(ds, coords['lon'], coords['lat'], resampling=Resampling.average)

    dif_attrs = dict(long_name='Surface Diffuse Shortwave Flux', units='W m-2')
    ds['influx_diffuse'] = (ds['SIS'] - ds['SID']) .assign_attrs(**dif_attrs)
    ds = ds.rename({'SID': 'influx_direct'}).drop_vars('SIS')
    ds = ds.assign_coords(x=ds.coords['lon'], y=ds.coords['lat'])

    return ds.swap_dims({'lon': 'x', 'lat':'y'})



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Functions for aggregating results.
"""

import xarray as xr
import dask

def aggregate_matrix(da, matrix, index):
    if index.name is None:
        index = index.rename('dim_0')
    if isinstance(da.data, dask.array.core.Array):
        da = da.stack(spatial=('y', 'x'))
        return xr.apply_ufunc(
            lambda da: da * matrix.T,
            da,
            input_core_dims=[['spatial']],
            output_core_dims=[[index.name]],
            dask='parallelized',
            output_dtypes=[da.dtype],
            dask_gufunc_kwargs = dict(output_sizes={index.name: index.size})
        ).assign_coords(**{index.name: index})
    else:
        da = da.stack(spatial=('y', 'x')).transpose('spatial', 'time')
        return xr.DataArray(matrix * da, [index, da.coords['time']])



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Module involving hydro operations in Atlite.
"""

import xarray as xr
import geopandas as gpd
import pandas as pd
import numpy as np

from collections import namedtuple
from shapely.geometry import Point
from tqdm import tqdm

import logging
logger = logging.getLogger(__name__)

Basins = namedtuple('Basins', ['plants', 'meta', 'shapes'])


def find_basin(shapes, lon, lat):
    hids = shapes.index[shapes.intersects(Point(lon, lat))]
    if len(hids) > 1:
        logger.warning(
            f"The point ({lon}, {lat}) is in several basins: {hids}. "
            "Assuming the first one.")
    return hids[0]


def find_upstream_basins(meta, hid):
    hids = [hid]
    i = 0
    while i < len(hids):
        hids.extend(meta.index[meta['NEXT_DOWN'] == hids[i]])
        i += 1
    return hids


def determine_basins(plants, hydrobasins, show_progress=True):
    if isinstance(hydrobasins, str):
        hydrobasins = gpd.read_file(hydrobasins)

    assert isinstance(hydrobasins, gpd.GeoDataFrame), (
        "hydrobasins should be passed as a filename or a GeoDataFrame, "
        "but received `type(hydrobasins) = {}`".format(type(hydrobasins))
    )

    missing_columns = (pd.Index(['HYBAS_ID',
                                 'DIST_MAIN',
                                 'NEXT_DOWN',
                                 'geometry']) .difference(hydrobasins.columns))
    assert missing_columns.empty, (
        "Couldn't find the column(s) {} in the hydrobasins dataset."
        .format(", ".join(missing_columns))
    )

    hydrobasins = hydrobasins.set_index("HYBAS_ID")

    meta = hydrobasins[hydrobasins.columns.difference(('geometry',))]
    shapes = hydrobasins['geometry']


    plant_basins = []
    for p in tqdm(plants.itertuples(), disable=not show_progress,
                  desc="Determine upstream basins per plant"):
        hid = find_basin(shapes, p.lon, p.lat)
        plant_basins.append((hid, find_upstream_basins(meta, hid)))
    plant_basins = pd.DataFrame(
        plant_basins, columns=[
            'hid', 'upstream'], index=plants.index)

    unique_basins = pd.Index(
        plant_basins['upstream'].sum()).unique().rename("hid")
    return Basins(
        plant_basins,
        meta.loc[unique_basins],
        shapes.loc[unique_basins])


def shift_and_aggregate_runoff_for_plants(
        basins, runoff, flowspeed=1, show_progress=True):
    inflow = xr.DataArray(np.zeros((len(basins.plants), runoff.indexes["time"].size)),
                          [('plant', basins.plants.index),
                           ('time', runoff.coords["time"])])

    for ppl in tqdm(basins.plants.itertuples(), disable=not show_progress,
                    desc="Shift and aggregate runoff by plant"):
        inflow_plant = inflow.loc[dict(plant=ppl.Index)]
        distances = (basins.meta.loc[ppl.upstream, "DIST_MAIN"] -
                     basins.meta.at[ppl.hid, "DIST_MAIN"])
        nhours = (distances * (flowspeed * 3.6) + 0.5).astype(int)

        for b in ppl.upstream:
            inflow_plant += runoff.sel(hid=b).shift(time=nhours.at[b])

    return inflow
